import pytest

from snowflake.core.database import Database
from snowflake.core.database_role import DatabaseRole
from snowflake.core.exceptions import NotFoundError
from tests.integ.utils import backup_database_and_schema
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


def test_drop(databases, cursor):
    try:
        with backup_database_and_schema(cursor):
            new_db_def = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
            new_db_def.comment = "database first"
            database_handle = databases.create(new_db_def)
            cursor.execute(f"USE DATABASE {new_db_def.name}")

            database_role_name = random_string(10, "test_database_role")
            database_role = DatabaseRole(name=database_role_name, comment="test_comment")

            database_role_handles = []

            database_roles = databases[new_db_def.name].database_roles

            try:
                # check there are no database roles yet
                assert len(list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))) == 0

                # happy path
                database_role_handles += [database_roles.create(database_role)]
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1
                assert database_roles_list[0].name == database_role_name.upper()
                assert database_roles_list[0].comment == "test_comment"

                # drop role
                database_roles[database_role_name].drop()
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 0

                # drop a random role
                with pytest.raises(NotFoundError):
                    database_roles["RANDOM"].drop()

                # drop with ifexists
                database_roles["RANDOM"].drop(if_exists=True)

            finally:
                for database_role_handle in database_role_handles:
                    database_role_handle.drop(if_exists=True)
    finally:
        database_handle.drop()



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.

import os

from collections.abc import Iterable, Iterator
from concurrent.futures import Future
from datetime import datetime, timedelta
from logging import getLogger
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Literal, NamedTuple, Optional, Union, overload

from pydantic import StrictStr

from snowflake.core import PollingOperation
from snowflake.core._common import CreateMode, SchemaObjectCollectionParent, SchemaObjectReferenceMixin
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._internal.utils import deprecated
from snowflake.core._operation import PollingOperations
from snowflake.core._options import require_snowpark
from snowflake.core.task._generated import (
    CronSchedule,
    MinutesSchedule,
    SuccessResponse,
    TagAssignment,
    TagReference,
    TaskApi,
    TaskRun,
    TaskSchedule,
)
from snowflake.core.task._generated.models import Task as TaskModel

from .._options import snowpark
from .._utils import tag_assignment_to_tag_tuple, tag_resource_to_tag_reference, tag_tuple_to_tag_assignment
from ..tag import TagResource
from ..tag._tag import TagValue


if TYPE_CHECKING:
    from snowflake.core.schema._schema import SchemaResource
    from snowflake.snowpark.stored_procedure import StoredProcedure
    from snowflake.snowpark.types import DataType

_logger = getLogger(__name__)


TASK_CONTEXT_FILE_IMPORT = (
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "context.py"),
    "snowflake.core.task.context",
)


class Cron(NamedTuple):
    """Specifies a cron expression and time zone for periodically running the task.

    Supports a subset of standard cron utility syntax.

    Examples
    ________

    >>> cron1 = Cron("0 0 10-20 * TUE,THU", "America/Los_Angeles")
    """

    expr: str
    """The cron expression. The minimum interval is 1 minute.

     It consists of the following fields:

    .. code-block::

        # __________ minute (0-59)
        # | ________ hour (0-23)
        # | | ______ day of month (1-31, or L)
        # | | | ____ month (1-12, JAN-DEC)
        # | | | | _ day of week (0-6, SUN-SAT, or L)
        # | | | | |
        # | | | | |
          * * * * *

    The following special characters are supported:

        - ``*`` Wildcard. Specifies any occurrence of the field.

        - ``L`` Stands for “last”. When used in the day-of-week field, it allows you to specify constructs such as
          “the last Friday” (“5L”) of a given month. In the day-of-month field, it specifies the last day of the month.

        - ``/n`` Indicates the nth instance of a given unit of time. Each quanta of time is computed independently.
          For example, if 4/3 is specified in the month field, then the task is scheduled for April, July and October
          (i.e. every 3 months, starting with the 4th month of the year). The same schedule is maintained in subsequent
          years. That is, the task is not scheduled to run in January (3 months after the October run).

    Tasks scheduled during specific times on days when the transition from standard time to daylight saving time
    (or the reverse) occurs can have unexpected behaviors. For example:

      - During the autumn change from daylight saving time to standard time, a task scheduled to start at 1 AM in
        the America/Los_Angeles time zone (i.e. 0 1 * * * America/Los_Angeles) would run twice: once at 1 AM and then
        again when 1:59:59 AM shifts to 1:00:00 AM local time. That is, there are two points in time when the local
        time is 1 AM.
      - During the spring change from standard time to daylight saving time, a task scheduled to start at 2 AM in
        the America/Los_Angeles time zone (i.e. 0 2 * * * America/Los_Angeles) would not run at all because the
        local time shifts from 1:59:59 AM to 3:00:00 AM. That is, there is no point during that day when the local
        time is 2 AM.

    To avoid unexpected task executions due to daylight saving time, use one of the following:
      - Do not schedule tasks to run at a specific time between 1 AM and 3 AM (daily, or on days of the week that
        include Sundays), or
      - Manually adjust the cron expression for tasks scheduled during those hours twice each year to compensate for
        the time change due to daylight saving time, or
      - Use a time format that does not apply daylight savings time, such as UTC.
    """
    timezone: str
    """The timezone for the cron expression.

    For a list of time zones, see the `list of tz database time zones in Wikipedia <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_.
    """

    def __eq__(self, other: "Cron") -> bool:  # type: ignore[override]
        return self.expr.lower() == other.expr.lower() and self.timezone.lower() == other.timezone.lower()


def _to_model_schedule(schedule: Optional[Union[Cron, timedelta]]) -> Optional[TaskSchedule]:
    if schedule is None:
        return None
    if isinstance(schedule, Cron):
        return CronSchedule(cron_expr=schedule.expr, timezone=schedule.timezone)
    elif isinstance(schedule, timedelta):
        total_seconds = schedule.total_seconds()
        if (not total_seconds.is_integer() or total_seconds % 60 != 0) or not (60 <= total_seconds <= 11520 * 60):
            raise ValueError(
                f"The schedule time delta must be a whole integer of minutes, greater than 1 minute and "
                f"less than 8 days (11520 minutes) but got {total_seconds / 60.0} minute(s)."
            )
        return MinutesSchedule(minutes=int(total_seconds / 60))
    raise TypeError("schedule should be either Cron or timedelta value")


def _to_model_target_completion_interval(target_completion_interval: Optional[timedelta]) -> Optional[MinutesSchedule]:
    if target_completion_interval is None:
        return None
    total_seconds = target_completion_interval.total_seconds()
    if (not total_seconds.is_integer() or total_seconds % 60 != 0) or not (60 <= total_seconds <= 24 * 60 * 60):
        raise ValueError(
            f"The target_completion_interval must be a whole integer of minutes with minimum of 1 minute and maximum "
            f"of 24 hours (1440 minutes) but got {total_seconds / 60.0} minute(s)."
        )
    return MinutesSchedule(minutes=int(total_seconds / 60))


def _from_model_schedule(schedule: Optional[TaskSchedule]) -> Optional[Union[timedelta, Cron]]:
    if schedule is None:
        return None
    if isinstance(schedule, MinutesSchedule):
        return timedelta(minutes=schedule.minutes)
    elif isinstance(schedule, CronSchedule):
        return Cron(schedule.cron_expr, schedule.timezone)
    raise TypeError("schedule must be either a MinutesSchedule or CronSchedule. ")  # won't happen in reality.


def _from_model_target_completion_interval(
    target_completion_interval: Optional[MinutesSchedule],
) -> Optional[timedelta]:
    if target_completion_interval is None:
        return None
    return timedelta(minutes=target_completion_interval.minutes)


class StoredProcedureCall:
    """Represents a procedure call used as a task's ``definition``.

    Parameters
    __________
    func: Union[Callable[..., Any], StoredProcedure]
        When it's a ``Callable``, typically a function, an anonymous stored procedure will be created as
        the Task's definition by using this ``Callable``.
        Note that the first parameter of your function should be a snowpark Session.

        When it's a ``StoredProcedure``, it will be converted to a SQL to call an existing stored procedure.
        The ``StoredProcedure`` must be a permanent one instead of a temporary one because a Task will run
        in a different session than the session that creates the Task. A temporary one won't be accessible
        from that session that runs the Task.
    args: list[Any], optional
        The arguments to call the stored procedure when ``func`` is a ``StoredProcedure``.
    return_type: DataType, optional
        A :class:`~snowflake.snowpark.types.DataType` representing the return data
        type of the stored procedure. Optional if type hints are provided.
    input_types: list[DataType], optional
        A list of :class:`~snowflake.snowpark.types.DataType`
        representing the input data types of the stored procedure. Optional if
        type hints are provided.
    stage_location: str, optional
        The stage location where the Python file for the anonymous stored procedure
        and its dependencies should be uploaded.
        It must be a permanent location because a Task will run in a different session than the session that
        creates the Task. A temporary one won't be accessible from that session that runs the Task.
    imports: list[Union[str, Tuple[str, str]]], optional
        A list of imports that only apply to this stored procedure. You can use a string to
        represent a file path (similar to the ``path`` argument in
        :meth:`~snowflake.snowpark.Session.add_import`) in this list, or a tuple of two
        strings to represent a file path and an import path (similar to the ``import_path``
        argument in :meth:`~snowflake.snowpark.Session.add_import`). These stored procedure-level imports
        will override the session-level imports added by
        :meth:`~snowflake.snowpark.Session.add_import`.
    packages: list[Union[str, ModuleType]], optional
        A list of packages that only apply to this stored procedure.
        These stored procedure-level packages will override the session-level packages added by
        :meth:`~snowflake.snowpark.Session.add_packages` and
        :meth:`~snowflake.snowpark.Session.add_requirements`.
    """

    def __init__(
        self,
        func: Union[Callable[..., Any], "StoredProcedure"],
        *,
        args: Optional[list[Any]] = None,
        return_type: Optional["DataType"] = None,
        input_types: Optional[list["DataType"]] = None,
        stage_location: Optional[str] = None,
        imports: Optional[list[Union[str, tuple[str, str]]]] = None,
        packages: Optional[list[Union[str, ModuleType]]] = None,
    ) -> None:
        require_snowpark()
        from snowflake.snowpark.stored_procedure import StoredProcedure

        if (not isinstance(func, StoredProcedure)) and stage_location is None:
            raise ValueError(
                "stage_location has to be specified when func is a Python function. And it must NOT be a temp location."
            )
        self.func: Union[Callable[..., Any], StoredProcedure] = func
        self._args = args if args else []
        self._return_type = return_type
        self._input_types = input_types
        self._stage_location = stage_location
        self._imports = imports
        self._packages = packages
        self._sql: Optional[str] = None


class Task:
    """Represents a Snowflake Task."""

    def __init__(
        self,
        name: str,
        definition: Union[str, StoredProcedureCall],
        *,
        warehouse: Optional[str] = None,
        user_task_managed_initial_warehouse_size: Optional[str] = None,
        target_completion_interval: Optional[timedelta] = None,
        serverless_task_min_statement_size: Optional[str] = None,
        serverless_task_max_statement_size: Optional[str] = None,
        suspend_task_after_num_failures: Optional[int] = None,
        user_task_timeout_ms: Optional[int] = None,
        schedule: Optional[Union[Cron, timedelta]] = None,
        allow_overlapping_execution: Optional[bool] = None,
        error_integration: Optional[str] = None,
        comment: Optional[str] = None,
        finalize: Optional[str] = None,
        task_auto_retry_attempts: Optional[int] = None,
        task_relations: Optional[str] = None,
        predecessors: Optional[list[str]] = None,
        condition: Optional[str] = None,
        config: Optional[dict[str, Any]] = None,
        session_parameters: Optional[dict[str, Any]] = None,
        id: Optional[str] = None,
        created_on: Optional[datetime] = None,
        last_committed_on: Optional[datetime] = None,
        last_suspended_on: Optional[datetime] = None,
        state: Optional[str] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        owner_role_type: Optional[str] = None,
    ) -> None:
        self.name: str = name  #: Name of the task.
        self.definition: Union[str, StoredProcedureCall] = definition
        """Definition of the task.

        Any one of the following:
          - A SQL Statement. It can be a single SQL statement, or a procedural logic using Snowflake Scripting.
          - A :class:`StoredProcedureCall` instance. This provides a Pythonic way to call an existing stored procedure,
            or use a Snowflake annonymous stored procedure in Python.
        """
        #: The virtual warehouse that provides compute resources for task runs. Omit this parameter if you want to use
        #: the serverless Snowflake-managed compute resources for runs of this task.
        self.warehouse = warehouse
        #: The size of the compute resources to provision for the first run of the task, before a task history is
        #: available for Snowflake to determine an ideal size. Once a task has successfully completed a few runs,
        #: Snowflake ignores this parameter setting. Note that if the task history is unavailable for a given task, the
        #: compute resources revert to this initial size.
        #:
        #: A warehouse size is the same as in `creating a virtual warehous
        #: <https://docs.snowflake.com/en/sql-reference/sql/create-warehouse>`_.
        #:
        #: If ``warehouse`` is specified for this task, then setting this parameter produces an error.
        self.user_task_managed_initial_warehouse_size = user_task_managed_initial_warehouse_size
        #: Specifies the desired task completion time. This parameter only applies to serverless tasks.
        #: This property is only set on a Task.
        self.target_completion_interval = target_completion_interval
        #: Specifies the minimum allowed warehouse size for the serverless task. This parameter only applies to
        #: serverless tasks. This parameter can be specified on the Task, Schema, Database, or Account.
        #: Precedence follows the standard parameter hierarchy
        self.serverless_task_min_statement_size = serverless_task_min_statement_size
        #: Specifies the maximum allowed warehouse size for the serverless task. This parameter only applies to
        #: serverless tasks. This parameter can be specified on the Task, Schema, Database, or Account.
        #: Precedence follows the standard parameter hierarchy
        self.serverless_task_max_statement_size = serverless_task_max_statement_size
        #: The schedule for periodically running the task. The minimum schedule is 1 minute.
        self.schedule = schedule
        #: Specifies the number of consecutive failed task runs after which the current task is suspended
        #: automatically. Failed task runs include runs in which task body either produces a user error or times
        #: out. Task runs that are skipped, canceled, or that fail due to a system error are considered indeterminate
        #: and are not included in the count of failed task runs.
        #:
        #: Set the parameter on a standalone task or the root task in a DAG. When the parameter is set to a value
        #: greater than 0, the following behavior applies to runs of the standalone task or DAG:
        #:
        #: Standalone tasks are automatically suspended after the specified number of consecutive task runs either fail
        #: or time out.
        #:
        #: The root task is automatically suspended after the run of any single task in a DAG fails or times out the
        #: specified number of times in consecutive runs.
        self.suspend_task_after_num_failures = suspend_task_after_num_failures
        #: Specifies the time limit on a single run of the task before it times out (in milliseconds).
        self.user_task_timeout_ms = user_task_timeout_ms
        #: Whether to allow multiple instances of the DAG to run concurrently.
        self.allow_overlapping_execution = allow_overlapping_execution
        #: Specifies the name of the notification integration used to communicate with Amazon SNS, MS Azure Event Grid,
        #: or Google Pub/Sub.
        #:
        #: For more information, refer to `Enabling Error Notifications for Tasks
        #: <https://docs.snowflake.com/en/user-guide/tasks-errors.html>`_.
        #:
        #: Required only when configuring a task to send error notifications using Amazon Simple Notification Service
        #: (SNS), Microsoft Azure Event Grid, or Google Pub/Sub.
        self.error_integration = error_integration
        #: Specifies a comment for the task.
        self.comment = comment
        #: Specifies the finalizer task, use this to add a finalizer task for the DAG. For more info
        # https://docs.snowflake.com/en/user-guide/tasks-intro#finalizer-task
        self.finalize = finalize
        #: Specifies the number of automatic task graph retry attempts. If any task graphs complete in a FAILED state,
        #: Snowflake can automatically retry the task graphs from the last task in the graph that failed.
        #:
        #: The automatic task graph retry is disabled by default. To enable this feature, set TASK_AUTO_RETRY_ATTEMPTS
        #: to a value greater than 0 on the root task of a task graph.
        #:
        #: Note that this parameter must be set to the root task of a task graph. If it’s set to a child task, an error
        #: will be returned.
        self.task_auto_retry_attempts = task_auto_retry_attempts
        #: Specifies the relationship between different task in the graph
        self.task_relations = task_relations
        #: Specifies one or more predecessor tasks for the current task. Use this option to create a DAG of tasks or add
        #: this task to an existing DAG.
        #:
        #: `A DAG <https://docs.snowflake.com/en/user-guide/tasks-intro#label-task-dag>`_ is a series of tasks that
        #: starts with a scheduled root task and is linked together by dependencies.
        self.predecessors = predecessors
        #: Specifies a Boolean SQL expression; multiple conditions joined with AND/OR are supported. When a task is
        #: triggered (based on its SCHEDULE or AFTER setting), it validates the conditions of the expression to
        #: determine whether to execute. If the conditions of the expression are not met, then the task skips the
        #: current run. Any tasks that identify this task as a predecessor also do not run.
        #:
        #: SYSTEM$STREAM_HAS_DATA is the only function supported for evaluation in the SQL expression. This function
        #: indicates whether a specified stream contains change tracking data.
        self.condition: Optional[str] = condition
        self.config = config
        """Set the configuration for the task. It can only be set on a root task then it applies to all tasks
        in the DAG.
        The parameter can be set on standalone tasks but does not affect the task behavior.
        Snowflake ensures only one instance of a standalone task is running at a given time.
        """

        self.session_parameters = session_parameters
        """Set the session parameters for the task at runtime."""

        # read-only properties

        #: Unique identifier for each task. Note that recreating a task essentially creates a new task, which has a new
        #  ID.
        self.id = id
        #: Date and time when the task was created.
        self.created_on = created_on
        #: Timestamp when a version of the task was last set. If no version has been set (i.e. if the task has not been
        #: resumed or manually executed after it was created), the value is NULL.
        self.last_committed_on = last_committed_on
        #: Timestamp when the task was last suspended. If the task has not been suspended yet, the value is NULL.
        self.last_suspended_on = last_suspended_on
        #: `"started"` or `"suspended"` based on the current state of the task.
        self.state = state
        #: Database in which the task is stored.
        self.database_name = database_name
        #: Schema in which the task is stored.
        self.schema_name = schema_name
        #: Role that owns the task (i.e. has the OWNERSHIP privilege on the task)
        self.owner = owner
        #: The type of role that owns the object, either ROLE or DATABASE_ROLE. Note that Snowflake returns NULL if you
        #: delete the object because there is no owner role for a deleted object.
        self.owner_role_type = owner_role_type

    def __repr__(self) -> str:
        return repr(self._to_rest_model())

    @property
    def sql_definition(self) -> str:
        """The definition of the task in SQL text.

        It's a readonly property. To set the ``definition`` of the ``Task``, use :attr:`definition`.

        If :attr:`definition` is a :class:`StoredProcedureCall`, the SQL that calls the stored procedure, or the
        anonymous stored procedure definition will be returned.
        """
        if isinstance(self.definition, str):
            return self.definition
        else:
            if self.definition._sql:
                return self.definition._sql
            else:
                raise ValueError("definition of this task can only be retrieved after creating the task")

    @classmethod
    def _from_rest_model(cls, model: TaskModel) -> "Task":
        return Task(
            name=model.name,
            definition=model.definition,
            warehouse=model.warehouse,
            suspend_task_after_num_failures=model.suspend_task_after_num_failures,
            user_task_managed_initial_warehouse_size=model.user_task_managed_initial_warehouse_size,
            target_completion_interval=_from_model_target_completion_interval(model.target_completion_interval),
            serverless_task_min_statement_size=model.serverless_task_min_statement_size,
            serverless_task_max_statement_size=model.serverless_task_max_statement_size,
            user_task_timeout_ms=model.user_task_timeout_ms,
            schedule=_from_model_schedule(model.schedule),
            allow_overlapping_execution=model.allow_overlapping_execution,
            error_integration=model.error_integration,
            comment=model.comment,
            finalize=model.finalize,
            task_auto_retry_attempts=model.task_auto_retry_attempts,
            task_relations=model.task_relations,
            predecessors=model.predecessors,
            condition=model.condition,
            config=model.config,
            session_parameters=model.session_parameters,
            id=model.id,
            created_on=model.created_on,
            last_committed_on=model.last_committed_on,
            last_suspended_on=model.last_suspended_on,
            state=model.state,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            owner_role_type=model.owner_role_type,
        )

    def _to_rest_model(self) -> TaskModel:
        for prop in ("config", "session_parameters"):
            attr_value = getattr(self, prop, None)
            if attr_value:
                for k, v in attr_value.items():
                    if not isinstance(v, (str, int, float, bool)):
                        raise TypeError(
                            f"Task.{prop} is a dict. The value of this dict must be one of str, int, float, or bool."
                            f"Found value type {type(v)} for key {k}"
                        )
        model = TaskModel(
            name=self.name,
            definition=self.sql_definition,
            warehouse=self.warehouse,
            user_task_managed_initial_warehouse_size=self.user_task_managed_initial_warehouse_size,
            target_completion_interval=_to_model_target_completion_interval(self.target_completion_interval),
            serverless_task_min_statement_size=self.serverless_task_min_statement_size,
            serverless_task_max_statement_size=self.serverless_task_max_statement_size,
            suspend_task_after_num_failures=self.suspend_task_after_num_failures,
            user_task_timeout_ms=self.user_task_timeout_ms,
            schedule=_to_model_schedule(self.schedule),
            allow_overlapping_execution=self.allow_overlapping_execution,
            error_integration=self.error_integration,
            comment=self.comment,
            finalize=self.finalize,
            task_auto_retry_attempts=self.task_auto_retry_attempts,
            task_relations=self.task_relations,
            predecessors=self.predecessors,
            condition=self.condition,
            config=self.config,
            session_parameters=self.session_parameters,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            owner_role_type=self.owner_role_type,
            id=self.id,
            state=self.state,
            created_on=self.created_on,
            last_committed_on=self.last_committed_on,
            last_suspended_on=self.last_suspended_on,
        )
        return model

    def to_dict(self, hide_readonly_properties: bool = False) -> dict[str, Any]:
        return self._to_rest_model().to_dict(hide_readonly_properties=hide_readonly_properties)


class TaskResource(SchemaObjectReferenceMixin["TaskCollection"]):
    """Represents a reference to a Snowflake Task resource.

    With this task reference, you can fetch information about a task, as well as perform certain
    actions on it.
    """

    def __init__(self, name: str, collection: "TaskCollection") -> None:
        self.collection = collection
        self.name = name

    @api_telemetry
    @deprecated("create_or_alter")
    def create_or_update(self, task: Task) -> None:
        self.create_or_alter(task=task)

    @api_telemetry
    def create_or_alter(self, task: Task) -> None:
        """Create a task in Snowflake or alter one if it already exists.

        The Snowflake task's properties will be updated to the properties of the input ``task`` if the task already
        exists.  Note that the full picture of a task is expected. If a property isn't set a value in the input
        ``task``, the property will be set to ``NULL`` in Snowflake too because it's regarded as the expected value.

        Parameters
        __________
        task: Task
            An instance of :class:`Task`.

        Examples
        ________
        >>> task_parameters = Task(name="your-task-name", definition="select 1")

        # Using a ``TaskCollection`` to create a reference to task in Snowflake server:

        >>> root.warehouses["your-task-name"].create_or_alter(task_parameters)
        """
        self._create_or_alter(task=task, async_req=False)

    @api_telemetry
    def create_or_alter_async(self, task: Task) -> PollingOperation[None]:
        """An asynchronous version of :func:`create_or_alter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._create_or_alter(task=task, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    @deprecated("drop")
    def delete(self) -> None:
        """Delete this task."""
        self.drop()

    @api_telemetry
    def drop(self, if_exists: Optional[bool] = None) -> None:
        """Drop this task.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this task before dropping it.
            Default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Deleting a task using its reference:

        >>> task_reference.drop()
        """
        self.collection._api.delete_task(
            self.database.name, self.schema.name, self.name, if_exists=if_exists, async_req=False
        )

    @api_telemetry
    def drop_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.delete_task(
            self.database.name, self.schema.name, self.name, if_exists=if_exists, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def fetch(self) -> Task:
        """Fetch the task resource.

        Examples
        ________
        Fetching a task using its reference:

        >>> task = task_reference.fetch()

        Accessing information of the task with task instance:

        >>> print(task.name, task.comment)
        """
        rest_model = self.collection._api.fetch_task(self.database.name, self.schema.name, self.name, async_req=False)
        return Task._from_rest_model(rest_model)

    @api_telemetry
    def fetch_async(self) -> PollingOperation[Task]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_task(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperation(future, lambda rest_model: Task._from_rest_model(rest_model))

    @api_telemetry
    def execute(self, *, retry_last: bool = False) -> None:
        """Execute the task immediately without waiting for the schedule.

        Parameters
        __________
        retry_last: bool, optional
            Re-execute the last failed task of the DAG. Default is ``False``.

        Examples
        ________
        Execute a task using its reference:

        >>> task_reference.execute()
        """
        self.collection._api.execute_task(
            self.database.name, self.schema.name, self.name, retry_last=retry_last, async_req=False
        )

    @api_telemetry
    def execute_async(self, *, retry_last: bool = False) -> PollingOperation[None]:
        """An asynchronous version of :func:`execute`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.execute_task(
            self.database.name, self.schema.name, self.name, retry_last=retry_last, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def resume(self) -> None:
        """Resume the task then it will run on the schedule.

        Examples
        ________
        Resume a task using its reference:

        >>> task_reference.resume()
        """
        self.collection._api.resume_task(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def resume_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`resume`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.resume_task(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def suspend(self) -> None:
        """Suspend the task so it won't run again on the schedule.

        Examples
        ________
        Suspend a task using its reference:

        >>> task_reference.suspend()
        """
        self.collection._api.suspend_task(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def suspend_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`suspend`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.suspend_task(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def fetch_task_dependents(self) -> list[Task]:
        """Return the list of child tasks that use this task as the root in a DAG.

        Examples
        ________
        Fetching the child tasks of a task using its reference:

        >>> child_tasks = task_reference.fetch_task_dependents()
        """
        return [
            Task._from_rest_model(x)
            for x in self.collection._api.fetch_task_dependents(
                self.database.name, self.schema.name, self.name, async_req=False
            )
        ]

    @api_telemetry
    def fetch_task_dependents_async(self) -> PollingOperation[list[Task]]:
        """An asynchronous version of :func:`fetch_task_dependents`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_task_dependents(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperation(future, lambda rest_models: [Task._from_rest_model(x) for x in rest_models])

    @api_telemetry
    def get_complete_graphs(self, *, error_only: bool = True) -> Iterable[TaskRun]:
        """Return the status of a completed graph run.

        It returns details for runs that executed successfully, failed, or were cancelled in the past 60 minutes.

        To retrieve the details for graph runs that are currently executing, or are next scheduled to run within the
        next 8 days, use :meth:`get_current_graphs`.

        Parameters
        __________
        error_only: bool, optional
            Return only the graph runs that have failed. Default is ``True``.

        Examples
        ________
        Getting the completed graph runs of a task using its reference:

        >>> completed_graphs = task_reference.get_complete_graphs()
        """
        return self.collection._api.get_complete_graphs(
            self.database.name, self.schema.name, self.name, error_only=error_only, async_req=False
        )

    @api_telemetry
    def get_complete_graphs_async(self, *, error_only: bool = True) -> PollingOperation[Iterable[TaskRun]]:
        """An asynchronous version of :func:`get_complete_graphs`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_complete_graphs(
            self.database.name, self.schema.name, self.name, error_only=error_only, async_req=True
        )
        return PollingOperations.identity(future)

    @api_telemetry
    def get_current_graphs(self) -> Iterable[TaskRun]:
        """Return the status of a graph run that is currently scheduled or is executing.

        It returns details for graph runs that are currently executing or are next scheduled to run within the next 8
        days.  To retrieve the details for graph runs that have completed in the past 60 minutes, use
        :meth:`get_complete_graphs`.

        Examples
        ________
        Getting the current graph runs of a task using its reference:

        >>> current_graphs = task_reference.get_current_graphs()
        """
        return self.collection._api.get_current_graphs(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def get_current_graphs_async(self) -> PollingOperation[Iterable[TaskRun]]:
        """An asynchronous version of :func:`get_current_graphs`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_current_graphs(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperations.identity(future)

    @api_telemetry
    def set_tags(self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None) -> None:
        """Set tags on a task.

        Parameters
        __________
        tags: dict[TagResource, TagValue]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        tag_assignments = [
            tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
            for [tag_resource, tag_value] in tags.items()
        ]

        self.collection._api.set_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_assignment=tag_assignments,
            if_exists=if_exists,
        )

    @api_telemetry
    def set_tags_async(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`set_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        tag_assignments = [
            tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
            for [tag_resource, tag_value] in tags.items()
        ]

        future = self.collection._api.set_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_assignment=tag_assignments,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def unset_tags(self, tag_resources: set[TagResource], if_exists: Optional[bool] = None) -> None:
        """Unset tags from a task.

        Parameters
        __________
        tag_resources: set[TagResource]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        tag_reference = [tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources]

        self.collection._api.unset_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_reference=tag_reference,
            if_exists=if_exists,
        )

    @api_telemetry
    def unset_tags_async(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`unset_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        tag_reference = [tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources]

        future = self.collection._api.unset_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_reference=tag_reference,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def get_tags(self, with_lineage: Optional[bool] = None) -> dict[TagResource, TagValue]:
        """Get the tag assignments for a task.

        Returns all tags assigned to a task. This operation requires an active warehouse.

        Parameters
        __________
        with_lineage: bool, optional
            Parameter that specifies whether tag assignments inherited by the object from its ancestors in securable
            object hierarchy should be returned as well: - `true`: All tags assigned to this object should be returned,
            inheritance included. - `false`: Only tags explicitly assigned to this object should be returned.
        """
        tag_assignments = self.collection._api.get_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            with_lineage=with_lineage,
        )

        get_tags_dict = dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)
        return get_tags_dict

    @api_telemetry
    def get_tags_async(self, with_lineage: Optional[bool] = None) -> PollingOperation[dict[TagResource, TagValue]]:
        """An asynchronous version of :func:`get_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            with_lineage=with_lineage,
            async_req=True,
        )
        return PollingOperation(
            future, lambda tag_assignments: dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)
        )

    @overload
    def _create_or_alter(self, task: Task, async_req: Literal[True]) -> Future[SuccessResponse]: ...

    @overload
    def _create_or_alter(self, task: Task, async_req: Literal[False]) -> SuccessResponse: ...

    def _create_or_alter(self, task: Task, async_req: bool) -> Union[SuccessResponse, Future[SuccessResponse]]:
        self.collection._extract_definition(task)
        task_model = task._to_rest_model()
        task_model.schedule = _to_model_schedule(task.schedule)
        return self.collection._api.create_or_alter_task(
            database=self.database.name,
            var_schema=self.schema.name,
            task=task_model,
            name=task.name,
            async_req=async_req,
        )


class TaskCollection(SchemaObjectCollectionParent[TaskResource]):
    """Represents the collection operations of the Snowflake Task resource.

    With this collection, you can create, iterate through, and search for task that you have access to
    in the current context.

    Examples
    ________
    >>> task_collection = root.databases["mydb"].schemas["myschema"].tasks
    >>> task = Task(name="mytask", definition="select 1")
    >>> task_collection.create(task)
    """

    def __init__(self, schema: "SchemaResource") -> None:
        super().__init__(schema, TaskResource)
        self._api = TaskApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(self, task: Task, *, mode: CreateMode = CreateMode.error_if_exists) -> TaskResource:
        """Create a task in Snowflake.

        Parameters
        __________
        task: an instance of :class:`Task`.
        mode: CreateMode, optional
            One of the following strings.

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the task already exists in Snowflake. Equivalent to SQL ``create task <name> ...``.

            ``CreateMode.or_replace``: Replace if the task already exists in Snowflake. Equivalent to SQL
            ``create or replace task <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the task already exists in Snowflake.
            Equivalent to SQL ``create task <name> if not exists...``

            Default value is ``CreateMode.error_if_exists``.

        Examples
        ________
        Creating a task in Snowflake and getting a reference to it:

        >>> task_parameters = Task(name="mytask", definition="select 1")
        >>> # Use the task collection created before to create a reference to the task resource
        >>> # in Snowflake.
        >>> task_reference = task_collection.create(task_parameters)
        """
        self._extract_definition(task)
        task_model = task._to_rest_model()
        real_mode = CreateMode[mode].value
        self._api.create_task(self.database.name, self.schema.name, task_model, real_mode, async_req=False)
        return self[task.name]

    @api_telemetry
    def create_async(
        self, task: Task, *, mode: CreateMode = CreateMode.error_if_exists
    ) -> PollingOperation[TaskResource]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        self._extract_definition(task)
        task_model = task._to_rest_model()
        real_mode = CreateMode[mode].value
        future = self._api.create_task(self.database.name, self.schema.name, task_model, real_mode, async_req=True)
        return PollingOperation(future, lambda _: self[task.name])

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
        root_only: bool = False,
    ) -> Iterator[Task]:
        """Iterate through ``Task`` objects in Snowflake, filtering on any optional ``like`` pattern.

        Parameters
        __________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).
        starts_with: str, optional
            String used to filter the command output based on the string of characters t
        limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.
        from_name: str, optional
            Fetch rows only following the first row whose object name matches
            the specified string. This is case-sensitive and does not have to be the full name.
        root_only: bool, optional
            Look for root tasks only. Default is ``False``.

        Examples
        ________
        Showing all tasks that you have access to see:

        >>> tasks = task_collection.iter()

        Showing information of the exact task you want to see:

        >>> tasks = task_collection.iter(like="your-task-name")

        Showing tasks starting with 'your-task-name-':

        >>> tasks = task_collection.iter(like="your-task-name-%")

        Using a for loop to retrieve information from iterator:

        >>> for task in tasks:
        ...     print(task.name, task.comment)
        """
        tasks = self._api.list_tasks(
            self.database.name,
            self.schema.name,
            root_only,
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            show_limit=limit,
            from_name=from_name,
            async_req=False,
        )

        return map(Task._from_rest_model, iter(tasks))

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
        root_only: bool = False,
    ) -> PollingOperation[Iterator[Task]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.list_tasks(
            self.database.name,
            self.schema.name,
            root_only,
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            show_limit=limit,
            from_name=from_name,
            async_req=True,
        )
        return PollingOperation(future, lambda tasks: map(Task._from_rest_model, iter(tasks)))

    def _extract_definition(self, task: Task) -> None:
        definition = task.definition
        if not isinstance(definition, str):
            require_snowpark()
            from snowflake.snowpark.stored_procedure import StoredProcedure

            if isinstance(definition.func, StoredProcedure):
                sproc_obj = definition.func
            else:
                imports = definition._imports if definition._imports else []
                if not snowpark._internal.utils.is_in_stored_procedure():
                    imports.append(TASK_CONTEXT_FILE_IMPORT)
                sproc_obj = self.root.session.sproc.register(
                    definition.func,
                    name="task_handler_sp",
                    return_type=definition._return_type,
                    input_types=definition._input_types,
                    stage_location=definition._stage_location,
                    imports=imports,
                    packages=definition._packages,
                    anonymous=True,
                    is_permanent=True,
                )
            sp_sql = snowpark._internal.udf_utils.generate_call_python_sp_sql(
                self.root.session, sproc_obj.name, *definition._args
            )
            if sproc_obj._anonymous_sp_sql:
                sp_sql = f"{sproc_obj._anonymous_sp_sql}{sp_sql}"
            definition._sql = sp_sql



"""
Cortex Search REST API.

OpenAPI 3.0 specification for the Cortex Search REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr

from snowflake.core.cortex.search_service._generated.models.scheduling_state import (
    SchedulingState,
)
from snowflake.core.cortex.search_service._generated.models.target_lag import TargetLag, TargetLagModel


class CortexSearchService(BaseModel):
    """A model object representing the CortexSearchService resource.

    Constructs an object of type CortexSearchService with the provided properties.

    Parameters
    __________
    name : str, optional
        Specifies the name for the cortex search service, must be unique for the schema in which the cortex search service is created
    search_column : str, optional
        Specifies the name of the search column for the cortex search service.
    columns : list[str], optional
        Specifies all columns included in the cortex search service and that can be returned in search queries.
    attribute_columns : list[str], optional
        Specifies the attribute columns, which can be referenced in filters in search queries to the cortex search service.
    target_lag : TargetLag, optional

    warehouse : str, optional
        Specifies the name of the warehouse that provides the compute resources for refreshing the cortex search service
    definition : str, optional
        Specifies the definition (source query) used to create the cortex search service
    comment : str, optional
        Specifies a comment for the cortex search service
    created_on : datetime, optional
        Date and time when the cortex search service was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the cortex search service is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the cortex search service is stored — **Read-only:** *any user-provided value will be ignored.*
    source_data_num_rows : int, optional
        Number of rows in the materialized source data feeding into the cortex search service — **Read-only:** *any user-provided value will be ignored.*
    data_timestamp : datetime, optional
        Date and time as of which data existent in base tables is now serving — **Read-only:** *any user-provided value will be ignored.*
    indexing_state : SchedulingState, optional

    serving_state : SchedulingState, optional

    indexing_error : str, optional
        Error encountered during the latest indexing pipeline of the cortex search service, if any — **Read-only:** *any user-provided value will be ignored.*
    serving_data_bytes : int, optional
        Size of the serving index, in bytes — **Read-only:** *any user-provided value will be ignored.*
    """

    name: Optional[StrictStr] = None

    search_column: Optional[StrictStr] = None

    columns: Optional[List[StrictStr]] = None

    attribute_columns: Optional[List[StrictStr]] = None

    target_lag: Optional[TargetLag] = None

    warehouse: Optional[StrictStr] = None

    definition: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    created_on: Optional[datetime] = None

    database_name: Optional[StrictStr] = None

    schema_name: Optional[StrictStr] = None

    source_data_num_rows: Optional[StrictInt] = None

    data_timestamp: Optional[datetime] = None

    indexing_state: Optional[SchedulingState] = None

    serving_state: Optional[SchedulingState] = None

    indexing_error: Optional[StrictStr] = None

    serving_data_bytes: Optional[StrictInt] = None

    __properties = [
        "name",
        "search_column",
        "columns",
        "attribute_columns",
        "target_lag",
        "warehouse",
        "definition",
        "comment",
        "created_on",
        "database_name",
        "schema_name",
        "source_data_num_rows",
        "data_timestamp",
        "indexing_state",
        "serving_state",
        "indexing_error",
        "serving_data_bytes",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CortexSearchService:
        """Create an instance of CortexSearchService from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "source_data_num_rows",
                    "data_timestamp",
                    "indexing_error",
                    "serving_data_bytes",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of target_lag
        if self.target_lag:
            _dict["target_lag"] = self.target_lag.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> CortexSearchService:
        """Create an instance of CortexSearchService from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return CortexSearchService.model_validate(obj)

        _obj = CortexSearchService.model_validate(
            {
                "name": obj.get("name"),
                "search_column": obj.get("search_column"),
                "columns": obj.get("columns"),
                "attribute_columns": obj.get("attribute_columns"),
                "target_lag": TargetLag.from_dict(obj.get("target_lag")) if obj.get("target_lag") is not None else None,
                "warehouse": obj.get("warehouse"),
                "definition": obj.get("definition"),
                "comment": obj.get("comment"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "source_data_num_rows": obj.get("source_data_num_rows"),
                "data_timestamp": obj.get("data_timestamp"),
                "indexing_state": obj.get("indexing_state"),
                "serving_state": obj.get("serving_state"),
                "indexing_error": obj.get("indexing_error"),
                "serving_data_bytes": obj.get("serving_data_bytes"),
            }
        )

        return _obj


class CortexSearchServiceModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        search_column: Optional[str] = None,
        columns: Optional[list[str]] = None,
        attribute_columns: Optional[list[str]] = None,
        target_lag: Optional[TargetLag] = None,
        warehouse: Optional[str] = None,
        definition: Optional[str] = None,
        comment: Optional[str] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        source_data_num_rows: Optional[int] = None,
        data_timestamp: Optional[datetime] = None,
        indexing_state: Optional[SchedulingState] = None,
        serving_state: Optional[SchedulingState] = None,
        indexing_error: Optional[str] = None,
        serving_data_bytes: Optional[int] = None,
    ):
        """A model object representing the CortexSearchService resource.

        Constructs an object of type CortexSearchService with the provided properties.

        Parameters
        __________
        name : str, optional
            Specifies the name for the cortex search service, must be unique for the schema in which the cortex search service is created
        search_column : str, optional
            Specifies the name of the search column for the cortex search service.
        columns : list[str], optional
            Specifies all columns included in the cortex search service and that can be returned in search queries.
        attribute_columns : list[str], optional
            Specifies the attribute columns, which can be referenced in filters in search queries to the cortex search service.
        target_lag : TargetLag, optional

        warehouse : str, optional
            Specifies the name of the warehouse that provides the compute resources for refreshing the cortex search service
        definition : str, optional
            Specifies the definition (source query) used to create the cortex search service
        comment : str, optional
            Specifies a comment for the cortex search service
        created_on : datetime, optional
            Date and time when the cortex search service was created.
        database_name : str, optional
            Database in which the cortex search service is stored
        schema_name : str, optional
            Schema in which the cortex search service is stored
        source_data_num_rows : int, optional
            Number of rows in the materialized source data feeding into the cortex search service.
        data_timestamp : datetime, optional
            Date and time as of which data existent in base tables is now serving.
        indexing_state : SchedulingState, optional

        serving_state : SchedulingState, optional

        indexing_error : str, optional
            Error encountered during the latest indexing pipeline of the cortex search service, if any.
        serving_data_bytes : int, optional
            Size of the serving index, in bytes.
        """
        self.name = name
        self.search_column = search_column
        self.columns = columns
        self.attribute_columns = attribute_columns
        self.target_lag = target_lag
        self.warehouse = warehouse
        self.definition = definition
        self.comment = comment
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.source_data_num_rows = source_data_num_rows
        self.data_timestamp = data_timestamp
        self.indexing_state = indexing_state
        self.serving_state = serving_state
        self.indexing_error = indexing_error
        self.serving_data_bytes = serving_data_bytes

    __properties = [
        "name",
        "search_column",
        "columns",
        "attribute_columns",
        "target_lag",
        "warehouse",
        "definition",
        "comment",
        "created_on",
        "database_name",
        "schema_name",
        "source_data_num_rows",
        "data_timestamp",
        "indexing_state",
        "serving_state",
        "indexing_error",
        "serving_data_bytes",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return CortexSearchService(
            name=self.name,
            search_column=self.search_column,
            columns=self.columns,
            attribute_columns=self.attribute_columns,
            target_lag=self.target_lag._to_model() if self.target_lag is not None else None,
            warehouse=self.warehouse,
            definition=self.definition,
            comment=self.comment,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            source_data_num_rows=self.source_data_num_rows,
            data_timestamp=self.data_timestamp,
            indexing_state=self.indexing_state,
            serving_state=self.serving_state,
            indexing_error=self.indexing_error,
            serving_data_bytes=self.serving_data_bytes,
        )

    @classmethod
    def _from_model(cls, model) -> CortexSearchServiceModel:
        return CortexSearchServiceModel(
            name=model.name,
            search_column=model.search_column,
            columns=model.columns,
            attribute_columns=model.attribute_columns,
            target_lag=TargetLagModel._from_model(model.target_lag) if model.target_lag else None,
            warehouse=model.warehouse,
            definition=model.definition,
            comment=model.comment,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            source_data_num_rows=model.source_data_num_rows,
            data_timestamp=model.data_timestamp,
            indexing_state=model.indexing_state,
            serving_state=model.serving_state,
            indexing_error=model.indexing_error,
            serving_data_bytes=model.serving_data_bytes,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a CortexSearchService.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the CortexSearchService object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> CortexSearchServiceModel:
        """Create an instance of CortexSearchService from a dict.

        This method constructs a CortexSearchService object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        CortexSearchService
            A CortexSearchService object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(CortexSearchService.from_dict(obj))


CortexSearchService._model_class = CortexSearchServiceModel



from os import PathLike, fspath
from typing import TYPE_CHECKING, Optional, Union

from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._internal.utils import deprecated, get_file, put_file
from snowflake.core.stage._generated.api.stage_api_base import StageCollectionBase, StageResourceBase


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class StageCollection(StageCollectionBase):
    """Represents the collection operations on the Snowflake Stage resource.

    With this collection, you can create, iterate through, and fetch stages
    that you have access to in the current context.

    Examples
    ________
    Creating a stage instance:

    >>> stages = root.databases["my_db"].schemas["my_schema"].stages
    >>> new_stage = Stage(name="my_stage", comment="This is a stage")
    >>> stages.create(new_stage)
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, StageResource)


class StageResource(StageResourceBase):
    """Represents a reference to a Snowflake stage.

    With this stage reference, you can fetch information about a stage, as well as perform
    certain actions on it.
    """

    _plural_name = "stages"

    def __init__(self, name: str, collection: StageCollection) -> None:
        super().__init__(name, collection)

    @api_telemetry
    @deprecated("drop")
    def delete(self) -> None:
        self.drop()

    @api_telemetry
    @deprecated("put")
    def upload_file(
        self, file_path: str, stage_folder_path: str, *, auto_compress: bool = True, overwrite: bool = False
    ) -> None:
        self.put(file_path, stage_folder_path, auto_compress=auto_compress, overwrite=overwrite)

    @api_telemetry
    @deprecated("get")
    def download_file(self, stage_path: str, file_folder_path: str) -> None:
        self.get(stage_path, file_folder_path)

    @api_telemetry
    def put(
        self,
        local_file_name: Union[str, PathLike],  # type: ignore[type-arg]
        stage_location: str,
        *,
        parallel: int = 4,
        auto_compress: bool = True,
        source_compression: str = "AUTO_DETECT",
        overwrite: bool = False,
    ) -> None:
        """Upload local files to a path in the stage.

        References: `Snowflake PUT command <https://docs.snowflake.com/en/sql-reference/sql/put.html>`_.

        Parameters
        __________
        local_file_name: str, ``PathLike``
            The path to the local files to upload. To match multiple files in the path,
            you can specify the wildcard characters ``*`` and ``?``.
        stage_location: str
            The prefix where you want to upload the files. e.g. ``/folder`` or ``/``
        parallel: int, optional
            Specifies the number of threads to use for uploading files. The upload process separates batches
            of data files by size:

            * Small files (< 64 MB) are staged in parallel as individual files.

            * Larger files are automatically split into chunks, staged concurrently, and reassembled
              in the target stage. A single thread can upload multiple chunks.

            Increasing the number of threads can improve performance when uploading large files.
            Supported values: Any integer value from 1 (no parallelism) to 99 (use 99 threads for uploading files).
        auto_compress: boolean, optional
            Specifies whether Snowflake uses gzip to compress files during upload.
            Default is ``True``.
        source_compression: str, optional
            Specifies the method of compression used on already-compressed files that are being staged.

            Values can be ``AUTO_DETECT``, ``GZIP``, ``BZ2``, ``BROTLI``, ``ZSTD``, ``DEFLATE``, ``RAW_DEFLATE``,
            ``NONE``, default is ``AUTO_DETECT``.
        overwrite: boolean, optional
            Specifies whether Snowflake will overwrite an existing file with the same name during upload.
            Default is ``False``.

        Examples
        ________
        Putting file on stage and compressing it using the stage's reference:

        >>> stage_reference.put("local_file.csv", "/folder", auto_compress=True)
        """
        stage_name = f"@{self.database.name}.{self.schema.name}.{self.name}"
        norm_stage_path = stage_location if stage_location.startswith("/") else f"/{stage_location}"
        put_file(
            self.collection.root,
            fspath(local_file_name),
            f"{stage_name}{norm_stage_path}",
            parallel=parallel,
            auto_compress=auto_compress,
            source_compression=source_compression,
            overwrite=overwrite,
        )

    @api_telemetry
    def get(
        self,
        stage_location: str,
        target_directory: Union[str, PathLike],  # type: ignore[type-arg]
        *,
        parallel: int = 4,
        pattern: Optional[str] = None,
    ) -> None:
        """Download the specified files from a path in the stage to a local directory.

        References: `Snowflake GET command <https://docs.snowflake.com/en/sql-reference/sql/get.html>`_.

        Parameters
        __________
        stage_location: str
            A directory or filename on a stage, from which you want to download the files.
            e.g. ``/folder/file_name.txt`` or ``/folder``
        target_directory: str, ``PathLike``
            The path to the local directory where the files should be downloaded.
            If ``target_directory`` does not already exist, the method creates the directory.
        parallel: int, optional
            Specifies the number of threads to use for downloading the files.
            The granularity unit for downloading is one file.
            Increasing the number of threads might improve performance when downloading large files.
            Valid values: Any integer value from 1 (no parallelism) to 99 (use 99 threads for downloading files).
        pattern: str, optional
            Specifies a regular expression pattern for filtering files to download.
            The command lists all files in the specified path and applies the regular expression pattern on each of
            the files found.
            Default: ``None`` (all files in the specified stage are downloaded).

        Examples
        ________
        Getting file from stage:

        >>> stage_reference.get("/folder/file_name.txt", "/local_folder")

        Getting files with a specific pattern:

        >>> stage_reference.get("/folder", "/local_folder", pattern=".*.txt")
        """
        stage_name = f"@{self.database.name}.{self.schema.name}.{self.name}"
        norm_stage_path = stage_location if stage_location.startswith("/") else f"/{stage_location}"
        get_file(
            self.collection.root,
            f"{stage_name}{norm_stage_path}",
            fspath(target_directory),
            parallel=parallel,
            pattern=pattern,
        )



"""
Snowflake Account API.

The Snowflake Account API is a REST API that you can use to access, update, and perform certain actions on Account resource in Snowflake.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, SecretStr, StrictBool, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class Account(BaseModel):
    """A model object representing the Account resource.

    Constructs an object of type Account with the provided properties.

    Parameters
    __________
    name : str
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    edition : str
        Snowflake Edition of the account. Possible values to use when creating an account include STANDARD, ENTERPRISE, and BUSINESS_CRITICAL.
    admin_name : str
        Name of the account administrator.
    email : str
        Email address of the account administrator.
    organization_name : str, optional
        Name of the organization — **Read-only:** *any user-provided value will be ignored.*
    region_group : str, optional
        Region group where the account is located. Note - This column is only displayed for organizations that span multiple region groups.
    region : str, optional
        Snowflake Region where the account is located. A Snowflake Region is a distinct location within a cloud platform region that is isolated from other Snowflake Regions. A Snowflake Region can be either multi-tenant or single-tenant (for a Virtual Private Snowflake account).
    created_on : datetime, optional
        Date and time the account was created — **Read-only:** *any user-provided value will be ignored.*
    account_url : str, optional
        Preferred Snowflake account URL that includes the values of organization_name and account_name — **Read-only:** *any user-provided value will be ignored.*
    account_locator : str, optional
        System-assigned identifier of the acccount — **Read-only:** *any user-provided value will be ignored.*
    account_locator_url : str, optional
        Legacy Snowflake account URL syntax that includes the region_name and account_locator — **Read-only:** *any user-provided value will be ignored.*
    managed_accounts : int, optional
        Indicates how many managed accounts have been created by the account — **Read-only:** *any user-provided value will be ignored.*
    consumption_billing_entity_name : str, optional
        Name of the consumption billing entity — **Read-only:** *any user-provided value will be ignored.*
    marketplace_consumer_billing_entity_name : str, optional
        Name of the marketplace consumer billing entity — **Read-only:** *any user-provided value will be ignored.*
    marketplace_provider_billing_entity_name : str, optional
        Name of the marketplace provider billing entity — **Read-only:** *any user-provided value will be ignored.*
    old_account_url : str, optional
        If the original account URL was saved when the account was renamed, provides the original URL. If the original account URL was dropped, the value is NULL even if the account was renamed — **Read-only:** *any user-provided value will be ignored.*
    comment : str, optional
        Optional comment in which to store information related to the account.
    is_org_admin : bool, optional
        Indicates whether the ORGADMIN role is enabled in an account. If TRUE, the role is enabled — **Read-only:** *any user-provided value will be ignored.*
    retention_time : int, optional
        Number of days that historical data is retained for Time Travel — **Read-only:** *any user-provided value will be ignored.*
    dropped_on : datetime, optional
        Date and time the account was dropped — **Read-only:** *any user-provided value will be ignored.*
    scheduled_deletion_time : datetime, optional
        Date and time when the account is scheduled to be permanently deleted. Accounts are deleted within one hour after the scheduled time — **Read-only:** *any user-provided value will be ignored.*
    restored_on : datetime, optional
        Date and time when the account was last restored — **Read-only:** *any user-provided value will be ignored.*
    account_old_url_saved_on : datetime, optional
        If the original account URL was saved when the account was renamed, provides the date and time when the original account URL was saved — **Read-only:** *any user-provided value will be ignored.*
    account_old_url_last_used : datetime, optional
        If the original account URL was saved when the account was renamed, indicates the last time the account was accessed using the original URL — **Read-only:** *any user-provided value will be ignored.*
    organization_old_url : str, optional
        If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the original account URL. If the original account URL was dropped, the value is NULL even if the organization changed — **Read-only:** *any user-provided value will be ignored.*
    organization_old_url_saved_on : datetime, optional
        If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the date and time when the original account URL was saved — **Read-only:** *any user-provided value will be ignored.*
    organization_old_url_last_used : datetime, optional
        If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, indicates the last time the account was accessed using the original account URL — **Read-only:** *any user-provided value will be ignored.*
    organization_url_expiration_on : datetime, optional
        If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the date and time when the original account URL will be dropped. Dropped URLs cannot be used to access the account — **Read-only:** *any user-provided value will be ignored.*
    moved_on : datetime, optional
        Date and time when the account was moved to a different organization — **Read-only:** *any user-provided value will be ignored.*
    is_events_account : bool, optional
        Indicates whether an account is an events account. For more information, see Set up logging and event sharing for an application — **Read-only:** *any user-provided value will be ignored.*
    moved_to_organization : str, optional
        If the account was moved to a different organization, provides the name of that organization — **Read-only:** *any user-provided value will be ignored.*
    admin_password : str, optional
        Password for the account administrator.
    admin_rsa_public_key : str, optional
        RSA public key for the account administrator.
    admin_user_type : str, optional
        User type of the account administrator.
    first_name : str, optional
        First name of the account administrator.
    last_name : str, optional
        Last name of the account administrator.
    must_change_password : bool,  default False
        Indicates whether the account administrator must change the password at the next login.
    polaris : bool,  default False
        Indicates whether the account is a Polaris account.
    """

    organization_name: Optional[StrictStr] = None

    name: Annotated[str, Field(strict=True)]

    region_group: Optional[StrictStr] = None

    region: Optional[StrictStr] = None

    edition: StrictStr

    created_on: Optional[datetime] = None

    account_url: Optional[StrictStr] = None

    account_locator: Optional[StrictStr] = None

    account_locator_url: Optional[StrictStr] = None

    managed_accounts: Optional[StrictInt] = None

    consumption_billing_entity_name: Optional[StrictStr] = None

    marketplace_consumer_billing_entity_name: Optional[StrictStr] = None

    marketplace_provider_billing_entity_name: Optional[StrictStr] = None

    old_account_url: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    is_org_admin: Optional[StrictBool] = None

    retention_time: Optional[StrictInt] = None

    dropped_on: Optional[datetime] = None

    scheduled_deletion_time: Optional[datetime] = None

    restored_on: Optional[datetime] = None

    account_old_url_saved_on: Optional[datetime] = None

    account_old_url_last_used: Optional[datetime] = None

    organization_old_url: Optional[StrictStr] = None

    organization_old_url_saved_on: Optional[datetime] = None

    organization_old_url_last_used: Optional[datetime] = None

    organization_url_expiration_on: Optional[datetime] = Field(default=None, alias="organization_URL_expiration_on")

    moved_on: Optional[datetime] = None

    is_events_account: Optional[StrictBool] = None

    moved_to_organization: Optional[StrictStr] = None

    admin_name: Optional[StrictStr]

    admin_password: Optional[SecretStr] = None

    admin_rsa_public_key: Optional[SecretStr] = None

    admin_user_type: Optional[StrictStr] = None

    first_name: Optional[StrictStr] = None

    last_name: Optional[StrictStr] = None

    email: Optional[StrictStr]

    must_change_password: Optional[StrictBool] = False

    polaris: Optional[StrictBool] = False

    __properties = [
        "organization_name",
        "name",
        "region_group",
        "region",
        "edition",
        "created_on",
        "account_url",
        "account_locator",
        "account_locator_url",
        "managed_accounts",
        "consumption_billing_entity_name",
        "marketplace_consumer_billing_entity_name",
        "marketplace_provider_billing_entity_name",
        "old_account_url",
        "comment",
        "is_org_admin",
        "retention_time",
        "dropped_on",
        "scheduled_deletion_time",
        "restored_on",
        "account_old_url_saved_on",
        "account_old_url_last_used",
        "organization_old_url",
        "organization_old_url_saved_on",
        "organization_old_url_last_used",
        "organization_URL_expiration_on",
        "moved_on",
        "is_events_account",
        "moved_to_organization",
        "admin_name",
        "admin_password",
        "admin_rsa_public_key",
        "admin_user_type",
        "first_name",
        "last_name",
        "email",
        "must_change_password",
        "polaris",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Account:
        """Create an instance of Account from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "organization_name",
                    "created_on",
                    "account_url",
                    "account_locator",
                    "account_locator_url",
                    "managed_accounts",
                    "consumption_billing_entity_name",
                    "marketplace_consumer_billing_entity_name",
                    "marketplace_provider_billing_entity_name",
                    "old_account_url",
                    "is_org_admin",
                    "retention_time",
                    "dropped_on",
                    "scheduled_deletion_time",
                    "restored_on",
                    "account_old_url_saved_on",
                    "account_old_url_last_used",
                    "organization_old_url",
                    "organization_old_url_saved_on",
                    "organization_old_url_last_used",
                    "organization_url_expiration_on",
                    "moved_on",
                    "is_events_account",
                    "moved_to_organization",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # set to None if dropped_on (nullable) is None
        if self.dropped_on is None:
            _dict["dropped_on"] = None

        # set to None if scheduled_deletion_time (nullable) is None
        if self.scheduled_deletion_time is None:
            _dict["scheduled_deletion_time"] = None

        # set to None if restored_on (nullable) is None
        if self.restored_on is None:
            _dict["restored_on"] = None

        # set to None if organization_url_expiration_on (nullable) is None
        if self.organization_url_expiration_on is None:
            _dict["organization_URL_expiration_on"] = None

        # set to None if moved_on (nullable) is None
        if self.moved_on is None:
            _dict["moved_on"] = None

        # set to None if admin_name (nullable) is None
        if self.admin_name is None:
            _dict["admin_name"] = None

        # set to None if email (nullable) is None
        if self.email is None:
            _dict["email"] = None

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Account:
        """Create an instance of Account from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Account.model_validate(obj)

        _obj = Account.model_validate(
            {
                "organization_name": obj.get("organization_name"),
                "name": obj.get("name"),
                "region_group": obj.get("region_group"),
                "region": obj.get("region"),
                "edition": obj.get("edition"),
                "created_on": obj.get("created_on"),
                "account_url": obj.get("account_url"),
                "account_locator": obj.get("account_locator"),
                "account_locator_url": obj.get("account_locator_url"),
                "managed_accounts": obj.get("managed_accounts"),
                "consumption_billing_entity_name": obj.get("consumption_billing_entity_name"),
                "marketplace_consumer_billing_entity_name": obj.get("marketplace_consumer_billing_entity_name"),
                "marketplace_provider_billing_entity_name": obj.get("marketplace_provider_billing_entity_name"),
                "old_account_url": obj.get("old_account_url"),
                "comment": obj.get("comment"),
                "is_org_admin": obj.get("is_org_admin"),
                "retention_time": obj.get("retention_time"),
                "dropped_on": obj.get("dropped_on"),
                "scheduled_deletion_time": obj.get("scheduled_deletion_time"),
                "restored_on": obj.get("restored_on"),
                "account_old_url_saved_on": obj.get("account_old_url_saved_on"),
                "account_old_url_last_used": obj.get("account_old_url_last_used"),
                "organization_old_url": obj.get("organization_old_url"),
                "organization_old_url_saved_on": obj.get("organization_old_url_saved_on"),
                "organization_old_url_last_used": obj.get("organization_old_url_last_used"),
                "organization_url_expiration_on": obj.get("organization_URL_expiration_on"),
                "moved_on": obj.get("moved_on"),
                "is_events_account": obj.get("is_events_account"),
                "moved_to_organization": obj.get("moved_to_organization"),
                "admin_name": obj.get("admin_name"),
                "admin_password": obj.get("admin_password"),
                "admin_rsa_public_key": obj.get("admin_rsa_public_key"),
                "admin_user_type": obj.get("admin_user_type"),
                "first_name": obj.get("first_name"),
                "last_name": obj.get("last_name"),
                "email": obj.get("email"),
                "must_change_password": obj.get("must_change_password")
                if obj.get("must_change_password") is not None
                else False,
                "polaris": obj.get("polaris") if obj.get("polaris") is not None else False,
            }
        )

        return _obj


class AccountModel:
    def __init__(
        self,
        name: str,
        edition: str,
        admin_name: str,
        email: str,
        # optional properties
        organization_name: Optional[str] = None,
        region_group: Optional[str] = None,
        region: Optional[str] = None,
        created_on: Optional[datetime] = None,
        account_url: Optional[str] = None,
        account_locator: Optional[str] = None,
        account_locator_url: Optional[str] = None,
        managed_accounts: Optional[int] = None,
        consumption_billing_entity_name: Optional[str] = None,
        marketplace_consumer_billing_entity_name: Optional[str] = None,
        marketplace_provider_billing_entity_name: Optional[str] = None,
        old_account_url: Optional[str] = None,
        comment: Optional[str] = None,
        is_org_admin: Optional[bool] = None,
        retention_time: Optional[int] = None,
        dropped_on: Optional[datetime] = None,
        scheduled_deletion_time: Optional[datetime] = None,
        restored_on: Optional[datetime] = None,
        account_old_url_saved_on: Optional[datetime] = None,
        account_old_url_last_used: Optional[datetime] = None,
        organization_old_url: Optional[str] = None,
        organization_old_url_saved_on: Optional[datetime] = None,
        organization_old_url_last_used: Optional[datetime] = None,
        organization_url_expiration_on: Optional[datetime] = None,
        moved_on: Optional[datetime] = None,
        is_events_account: Optional[bool] = None,
        moved_to_organization: Optional[str] = None,
        admin_password: Optional[str] = None,
        admin_rsa_public_key: Optional[str] = None,
        admin_user_type: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        must_change_password: Optional[bool] = False,
        polaris: Optional[bool] = False,
    ):
        """A model object representing the Account resource.

        Constructs an object of type Account with the provided properties.

        Parameters
        __________
        name : str
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        edition : str
            Snowflake Edition of the account. Possible values to use when creating an account include STANDARD, ENTERPRISE, and BUSINESS_CRITICAL.
        admin_name : str
            Name of the account administrator.
        email : str
            Email address of the account administrator.
        organization_name : str, optional
            Name of the organization.
        region_group : str, optional
            Region group where the account is located. Note - This column is only displayed for organizations that span multiple region groups.
        region : str, optional
            Snowflake Region where the account is located. A Snowflake Region is a distinct location within a cloud platform region that is isolated from other Snowflake Regions. A Snowflake Region can be either multi-tenant or single-tenant (for a Virtual Private Snowflake account).
        created_on : datetime, optional
            Date and time the account was created.
        account_url : str, optional
            Preferred Snowflake account URL that includes the values of organization_name and account_name.
        account_locator : str, optional
            System-assigned identifier of the acccount.
        account_locator_url : str, optional
            Legacy Snowflake account URL syntax that includes the region_name and account_locator.
        managed_accounts : int, optional
            Indicates how many managed accounts have been created by the account.
        consumption_billing_entity_name : str, optional
            Name of the consumption billing entity.
        marketplace_consumer_billing_entity_name : str, optional
            Name of the marketplace consumer billing entity.
        marketplace_provider_billing_entity_name : str, optional
            Name of the marketplace provider billing entity.
        old_account_url : str, optional
            If the original account URL was saved when the account was renamed, provides the original URL. If the original account URL was dropped, the value is NULL even if the account was renamed
        comment : str, optional
            Optional comment in which to store information related to the account.
        is_org_admin : bool, optional
            Indicates whether the ORGADMIN role is enabled in an account. If TRUE, the role is enabled.
        retention_time : int, optional
            Number of days that historical data is retained for Time Travel.
        dropped_on : datetime, optional
            Date and time the account was dropped.
        scheduled_deletion_time : datetime, optional
            Date and time when the account is scheduled to be permanently deleted. Accounts are deleted within one hour after the scheduled time.
        restored_on : datetime, optional
            Date and time when the account was last restored.
        account_old_url_saved_on : datetime, optional
            If the original account URL was saved when the account was renamed, provides the date and time when the original account URL was saved.
        account_old_url_last_used : datetime, optional
            If the original account URL was saved when the account was renamed, indicates the last time the account was accessed using the original URL.
        organization_old_url : str, optional
            If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the original account URL. If the original account URL was dropped, the value is NULL even if the organization changed.
        organization_old_url_saved_on : datetime, optional
            If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the date and time when the original account URL was saved.
        organization_old_url_last_used : datetime, optional
            If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, indicates the last time the account was accessed using the original account URL.
        organization_url_expiration_on : datetime, optional
            If the account’s organization was changed in a way that created a new account URL and the original account URL was saved, provides the date and time when the original account URL will be dropped. Dropped URLs cannot be used to access the account.
        moved_on : datetime, optional
            Date and time when the account was moved to a different organization.
        is_events_account : bool, optional
            Indicates whether an account is an events account. For more information, see Set up logging and event sharing for an application.
        moved_to_organization : str, optional
            If the account was moved to a different organization, provides the name of that organization.
        admin_password : str, optional
            Password for the account administrator.
        admin_rsa_public_key : str, optional
            RSA public key for the account administrator.
        admin_user_type : str, optional
            User type of the account administrator.
        first_name : str, optional
            First name of the account administrator.
        last_name : str, optional
            Last name of the account administrator.
        must_change_password : bool,  default False
            Indicates whether the account administrator must change the password at the next login.
        polaris : bool,  default False
            Indicates whether the account is a Polaris account.
        """
        self.organization_name = organization_name
        self.name = name
        self.region_group = region_group
        self.region = region
        self.edition = edition
        self.created_on = created_on
        self.account_url = account_url
        self.account_locator = account_locator
        self.account_locator_url = account_locator_url
        self.managed_accounts = managed_accounts
        self.consumption_billing_entity_name = consumption_billing_entity_name
        self.marketplace_consumer_billing_entity_name = marketplace_consumer_billing_entity_name
        self.marketplace_provider_billing_entity_name = marketplace_provider_billing_entity_name
        self.old_account_url = old_account_url
        self.comment = comment
        self.is_org_admin = is_org_admin
        self.retention_time = retention_time
        self.dropped_on = dropped_on
        self.scheduled_deletion_time = scheduled_deletion_time
        self.restored_on = restored_on
        self.account_old_url_saved_on = account_old_url_saved_on
        self.account_old_url_last_used = account_old_url_last_used
        self.organization_old_url = organization_old_url
        self.organization_old_url_saved_on = organization_old_url_saved_on
        self.organization_old_url_last_used = organization_old_url_last_used
        self.organization_url_expiration_on = organization_url_expiration_on
        self.moved_on = moved_on
        self.is_events_account = is_events_account
        self.moved_to_organization = moved_to_organization
        self.admin_name = admin_name
        self.admin_password = admin_password
        self.admin_rsa_public_key = admin_rsa_public_key
        self.admin_user_type = admin_user_type
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.must_change_password = must_change_password
        self.polaris = polaris

    __properties = [
        "organization_name",
        "name",
        "region_group",
        "region",
        "edition",
        "created_on",
        "account_url",
        "account_locator",
        "account_locator_url",
        "managed_accounts",
        "consumption_billing_entity_name",
        "marketplace_consumer_billing_entity_name",
        "marketplace_provider_billing_entity_name",
        "old_account_url",
        "comment",
        "is_org_admin",
        "retention_time",
        "dropped_on",
        "scheduled_deletion_time",
        "restored_on",
        "account_old_url_saved_on",
        "account_old_url_last_used",
        "organization_old_url",
        "organization_old_url_saved_on",
        "organization_old_url_last_used",
        "organization_URL_expiration_on",
        "moved_on",
        "is_events_account",
        "moved_to_organization",
        "admin_name",
        "admin_password",
        "admin_rsa_public_key",
        "admin_user_type",
        "first_name",
        "last_name",
        "email",
        "must_change_password",
        "polaris",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Account(
            organization_name=self.organization_name,
            name=self.name,
            region_group=self.region_group,
            region=self.region,
            edition=self.edition,
            created_on=self.created_on,
            account_url=self.account_url,
            account_locator=self.account_locator,
            account_locator_url=self.account_locator_url,
            managed_accounts=self.managed_accounts,
            consumption_billing_entity_name=self.consumption_billing_entity_name,
            marketplace_consumer_billing_entity_name=self.marketplace_consumer_billing_entity_name,
            marketplace_provider_billing_entity_name=self.marketplace_provider_billing_entity_name,
            old_account_url=self.old_account_url,
            comment=self.comment,
            is_org_admin=self.is_org_admin,
            retention_time=self.retention_time,
            dropped_on=self.dropped_on,
            scheduled_deletion_time=self.scheduled_deletion_time,
            restored_on=self.restored_on,
            account_old_url_saved_on=self.account_old_url_saved_on,
            account_old_url_last_used=self.account_old_url_last_used,
            organization_old_url=self.organization_old_url,
            organization_old_url_saved_on=self.organization_old_url_saved_on,
            organization_old_url_last_used=self.organization_old_url_last_used,
            organization_url_expiration_on=self.organization_url_expiration_on,
            moved_on=self.moved_on,
            is_events_account=self.is_events_account,
            moved_to_organization=self.moved_to_organization,
            admin_name=self.admin_name,
            admin_password=self.admin_password,
            admin_rsa_public_key=self.admin_rsa_public_key,
            admin_user_type=self.admin_user_type,
            first_name=self.first_name,
            last_name=self.last_name,
            email=self.email,
            must_change_password=self.must_change_password,
            polaris=self.polaris,
        )

    @classmethod
    def _from_model(cls, model) -> AccountModel:
        return AccountModel(
            organization_name=model.organization_name,
            name=model.name,
            region_group=model.region_group,
            region=model.region,
            edition=model.edition,
            created_on=model.created_on,
            account_url=model.account_url,
            account_locator=model.account_locator,
            account_locator_url=model.account_locator_url,
            managed_accounts=model.managed_accounts,
            consumption_billing_entity_name=model.consumption_billing_entity_name,
            marketplace_consumer_billing_entity_name=model.marketplace_consumer_billing_entity_name,
            marketplace_provider_billing_entity_name=model.marketplace_provider_billing_entity_name,
            old_account_url=model.old_account_url,
            comment=model.comment,
            is_org_admin=model.is_org_admin,
            retention_time=model.retention_time,
            dropped_on=model.dropped_on,
            scheduled_deletion_time=model.scheduled_deletion_time,
            restored_on=model.restored_on,
            account_old_url_saved_on=model.account_old_url_saved_on,
            account_old_url_last_used=model.account_old_url_last_used,
            organization_old_url=model.organization_old_url,
            organization_old_url_saved_on=model.organization_old_url_saved_on,
            organization_old_url_last_used=model.organization_old_url_last_used,
            organization_url_expiration_on=model.organization_url_expiration_on,
            moved_on=model.moved_on,
            is_events_account=model.is_events_account,
            moved_to_organization=model.moved_to_organization,
            admin_name=model.admin_name,
            admin_password=model.admin_password,
            admin_rsa_public_key=model.admin_rsa_public_key,
            admin_user_type=model.admin_user_type,
            first_name=model.first_name,
            last_name=model.last_name,
            email=model.email,
            must_change_password=model.must_change_password,
            polaris=model.polaris,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Account.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Account object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> AccountModel:
        """Create an instance of Account from a dict.

        This method constructs a Account object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Account
            A Account object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Account.from_dict(obj))


Account._model_class = AccountModel



from contextlib import suppress

from snowflake.core.dynamic_table import DownstreamLag, DynamicTable
from snowflake.core.exceptions import NotFoundError
from tests.utils import random_string


def test_swap(dynamic_tables, table_handle, db_parameters):
    table1_name = random_string(10, "test_table_")
    table2_name = random_string(10, "test_table_")

    test_table1 = DynamicTable(
        name=table1_name,
        warehouse=db_parameters["warehouse"],
        target_lag=DownstreamLag(),
        query=f"SELECT c1 FROM {table_handle.name}",
    )
    test_table2 = DynamicTable(
        name=table2_name,
        warehouse=db_parameters["warehouse"],
        target_lag=DownstreamLag(),
        query=f"SELECT c2 FROM {table_handle.name}",
    )

    try:
        test_table1_handle = dynamic_tables.create(test_table1)
        test_table2_handle = dynamic_tables.create(test_table2)

        test_table1_handle.swap_with(table2_name)
        dynamic_tables["dummy___table"].swap_with(table2_name, if_exists=True)

        fetched_table1 = test_table1_handle.fetch()
        fetched_table2 = test_table2_handle.fetch()

        assert fetched_table1.query == f"SELECT c2 FROM {table_handle.name}"
        assert fetched_table2.query == f"SELECT c1 FROM {table_handle.name}"
    finally:
        with suppress(NotFoundError):
            test_table1_handle.drop()
        with suppress(NotFoundError):
            test_table2_handle.drop()



"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class NonStreamingCompleteResponseChoicesInnerMessage(BaseModel):
    """A model object representing the NonStreamingCompleteResponseChoicesInnerMessage resource.

    Constructs an object of type NonStreamingCompleteResponseChoicesInnerMessage with the provided properties.

    Parameters
    __________
    content : str, optional
        The text completion response.
    content_list : list[object], optional
        Contents of text and toolUse response.
    """

    content: Optional[StrictStr] = None

    content_list: Optional[List[Any]] = None

    __properties = ["content", "content_list"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> NonStreamingCompleteResponseChoicesInnerMessage:
        """Create an instance of NonStreamingCompleteResponseChoicesInnerMessage from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> NonStreamingCompleteResponseChoicesInnerMessage:
        """Create an instance of NonStreamingCompleteResponseChoicesInnerMessage from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return NonStreamingCompleteResponseChoicesInnerMessage.model_validate(obj)

        _obj = NonStreamingCompleteResponseChoicesInnerMessage.model_validate(
            {
                "content": obj.get("content"),
                "content_list": obj.get("content_list"),
            }
        )

        return _obj


class NonStreamingCompleteResponseChoicesInnerMessageModel:
    def __init__(
        self,  # optional properties
        content: Optional[str] = None,
        content_list: Optional[list[object]] = None,
    ):
        """A model object representing the NonStreamingCompleteResponseChoicesInnerMessage resource.

        Constructs an object of type NonStreamingCompleteResponseChoicesInnerMessage with the provided properties.

        Parameters
        __________
        content : str, optional
            The text completion response.
        content_list : list[object], optional
            Contents of text and toolUse response.
        """
        self.content = content
        self.content_list = content_list

    __properties = ["content", "content_list"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return NonStreamingCompleteResponseChoicesInnerMessage(
            content=self.content,
            content_list=self.content_list,
        )

    @classmethod
    def _from_model(cls, model) -> NonStreamingCompleteResponseChoicesInnerMessageModel:
        return NonStreamingCompleteResponseChoicesInnerMessageModel(
            content=model.content,
            content_list=model.content_list,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a NonStreamingCompleteResponseChoicesInnerMessage.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the NonStreamingCompleteResponseChoicesInnerMessage object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> NonStreamingCompleteResponseChoicesInnerMessageModel:
        """Create an instance of NonStreamingCompleteResponseChoicesInnerMessage from a dict.

        This method constructs a NonStreamingCompleteResponseChoicesInnerMessage object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        NonStreamingCompleteResponseChoicesInnerMessage
            A NonStreamingCompleteResponseChoicesInnerMessage object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(NonStreamingCompleteResponseChoicesInnerMessage.from_dict(obj))


NonStreamingCompleteResponseChoicesInnerMessage._model_class = NonStreamingCompleteResponseChoicesInnerMessageModel



from contextlib import suppress

from snowflake.core.exceptions import NotFoundError

from ..utils import random_object_name


task_name1 = random_object_name()


def test_create_task_schedule_cron(tasks):
    from snowflake.core.task import Cron, Task

    try:
        schedule = Cron("0 9-17 * * SUN", "America/Los_Angeles")
        task = tasks.create(Task(name=task_name1, definition="select current_version()", schedule=schedule))
        assert task.fetch().schedule == schedule
        assert task.fetch().state == "suspended"
        task.suspend()
        assert task.fetch().state == "suspended"
        task.resume()
        assert task.fetch().state != "suspended"
        task.suspend()
        assert task.fetch().state == "suspended"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()



"""Manages Snowflake EventTables."""

from ..event_table._generated.models import EventTable, EventTableColumn
from ._event_table import EventTableCollection, EventTableResource


__all__ = ["EventTable", "EventTableResource", "EventTableCollection", "EventTableColumn"]



"""
Snowflake Event Table API.

The Snowflake Event Table API is a REST API that you can use to access, update, and perform certain actions on Event Table resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictBool, StrictStr


class EventTableColumn(BaseModel):
    """A model object representing the EventTableColumn resource.

    Constructs an object of type EventTableColumn with the provided properties.

    Parameters
    __________
    name : str, optional
        Column name
    datatype : str, optional
        The data type for the column
    nullable : bool, optional
        Specifies that the column does allow NULL values or not
    default : str, optional
        Specifies whether a default value is automatically inserted in the column if a value is not explicitly specified via an INSERT or CREATE TABLE AS SELECT statement
    primary_key : bool, optional
        A primary key is the column or columns that contain values that uniquely identify each row in a table
    unique_key : bool, optional
        Unique keys are columns in a table that uniquely identify items in the rows. This sounds a lot like a primary key, but the main difference is that unique keys can have NULL values.
    check : str, optional

    expression : str, optional

    comment : str, optional
        Specifies a comment for the column
    """

    name: Optional[StrictStr] = None

    datatype: Optional[StrictStr] = None

    nullable: Optional[StrictBool] = None

    default: Optional[StrictStr] = None

    primary_key: Optional[StrictBool] = None

    unique_key: Optional[StrictBool] = None

    check: Optional[StrictStr] = None

    expression: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    __properties = [
        "name",
        "datatype",
        "nullable",
        "default",
        "primary_key",
        "unique_key",
        "check",
        "expression",
        "comment",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> EventTableColumn:
        """Create an instance of EventTableColumn from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> EventTableColumn:
        """Create an instance of EventTableColumn from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return EventTableColumn.model_validate(obj)

        _obj = EventTableColumn.model_validate(
            {
                "name": obj.get("name"),
                "datatype": obj.get("datatype"),
                "nullable": obj.get("nullable"),
                "default": obj.get("default"),
                "primary_key": obj.get("primary_key"),
                "unique_key": obj.get("unique_key"),
                "check": obj.get("check"),
                "expression": obj.get("expression"),
                "comment": obj.get("comment"),
            }
        )

        return _obj


class EventTableColumnModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        datatype: Optional[str] = None,
        nullable: Optional[bool] = None,
        default: Optional[str] = None,
        primary_key: Optional[bool] = None,
        unique_key: Optional[bool] = None,
        check: Optional[str] = None,
        expression: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        """A model object representing the EventTableColumn resource.

        Constructs an object of type EventTableColumn with the provided properties.

        Parameters
        __________
        name : str, optional
            Column name
        datatype : str, optional
            The data type for the column
        nullable : bool, optional
            Specifies that the column does allow NULL values or not
        default : str, optional
            Specifies whether a default value is automatically inserted in the column if a value is not explicitly specified via an INSERT or CREATE TABLE AS SELECT statement
        primary_key : bool, optional
            A primary key is the column or columns that contain values that uniquely identify each row in a table
        unique_key : bool, optional
            Unique keys are columns in a table that uniquely identify items in the rows. This sounds a lot like a primary key, but the main difference is that unique keys can have NULL values.
        check : str, optional

        expression : str, optional

        comment : str, optional
            Specifies a comment for the column
        """
        self.name = name
        self.datatype = datatype
        self.nullable = nullable
        self.default = default
        self.primary_key = primary_key
        self.unique_key = unique_key
        self.check = check
        self.expression = expression
        self.comment = comment

    __properties = [
        "name",
        "datatype",
        "nullable",
        "default",
        "primary_key",
        "unique_key",
        "check",
        "expression",
        "comment",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return EventTableColumn(
            name=self.name,
            datatype=self.datatype,
            nullable=self.nullable,
            default=self.default,
            primary_key=self.primary_key,
            unique_key=self.unique_key,
            check=self.check,
            expression=self.expression,
            comment=self.comment,
        )

    @classmethod
    def _from_model(cls, model) -> EventTableColumnModel:
        return EventTableColumnModel(
            name=model.name,
            datatype=model.datatype,
            nullable=model.nullable,
            default=model.default,
            primary_key=model.primary_key,
            unique_key=model.unique_key,
            check=model.check,
            expression=model.expression,
            comment=model.comment,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a EventTableColumn.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the EventTableColumn object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> EventTableColumnModel:
        """Create an instance of EventTableColumn from a dict.

        This method constructs a EventTableColumn object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        EventTableColumn
            A EventTableColumn object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(EventTableColumn.from_dict(obj))


EventTableColumn._model_class = EventTableColumnModel



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.image_repository import ImageRepository, ImageRepositoryResource
from snowflake.core.image_repository._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def image_repositories(schema):
    return schema.image_repositories


@pytest.fixture
def image_repository(image_repositories):
    return image_repositories["my_rep"]


def test_create_image_repository(fake_root, image_repositories):
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories?createMode=errorIfExists",
    )
    kwargs = extra_params(query_params=[("createMode", "errorIfExists")], body={"name": "my_rep"})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        ir_res = image_repositories.create(ImageRepository(name="my_rep"))
        assert isinstance(ir_res, ImageRepositoryResource)
        assert ir_res.name == "my_rep"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = image_repositories.create_async(ImageRepository(name="my_rep"))
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_rep"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_image_repository(fake_root, image_repositories):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        image_repositories.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = image_repositories.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_image_repository(fake_root, image_repository):
    from snowflake.core.image_repository._generated.models import ImageRepository as ImageRepositoryModel

    model = ImageRepositoryModel(name="my_rep")
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        image_repository.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = image_repository.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == ImageRepository(name="my_rep").to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_image_repository(fake_root, image_repository):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        image_repository.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = image_repository.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_list_images_in_repository(fake_root, image_repository):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep/images")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        images = image_repository.list_images_in_repository()
        assert list(images) == []
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = image_repository.list_images_in_repository_async()
        assert isinstance(op, PollingOperation)
        images = op.result()
        assert list(images) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_set_tags(fake_root, image_repository, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        image_repository.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = image_repository.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, image_repository, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        image_repository.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = image_repository.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, image_repository):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/image-repositories/my_rep:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert image_repository.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = image_repository.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



import json
import sys

from enum import Enum, EnumMeta
from typing import Annotated, Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, StringConstraints


class CaseInsensitiveEnumMeta(EnumMeta):
    def __init__(cls, *args, **kws) -> None:  # type: ignore
        super().__init__(*args, **kws)

        class lookup(dict):  # type: ignore
            def get(self, key, default=None):  # type: ignore
                return super().get(key.lower(), key.lower())

        cls._legacy_mode_map_ = lookup({item.value.lower(): item.name for item in cls})  # type: ignore

    def __getitem__(cls, name: str) -> Any:
        converted_name = cls._legacy_mode_map_.get(name)
        return super().__getitem__(converted_name)


class CreateMode(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Enum for Snowflake create modes."""

    error_if_exists = "errorIfExists"
    or_replace = "orReplace"
    if_not_exists = "ifNotExists"


class DeleteMode(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Enum for delete modes."""

    cascade = "cascade"
    restrict = "restrict"


class Clone(BaseModel):
    source: Annotated[str, StringConstraints(strict=True)] = Field(...)
    point_of_time: Optional["PointOfTime"] = None


class PointOfTime(BaseModel):
    point_of_time_type: Annotated[str, StringConstraints(strict=True)] = Field(...)
    reference: Annotated[str, StringConstraints(strict=True)] = Field(
        ..., description="The relation to the point of time. At the time of writing at and before are supported."
    )
    when: Annotated[str, StringConstraints(strict=True)] = Field(
        ..., description="The actual description of the point of time."
    )
    __properties = ["point_of_time_type", "reference", "when"]

    __discriminator_property_name = "point_of_time_type"

    __discriminator_value_class_map = {
        "offset": "PointOfTimeOffset",
        "statement": "PointOfTimeStatement",
        "timestamp": "PointOfTimeTimestamp",
    }

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_dict(self) -> dict[str, str]:
        d = {p: getattr(self, p) for p in self.__properties}
        # Need to map "when" to the discriminator value as per our OAS
        d[d[self.__discriminator_property_name]] = d["when"]
        del d["when"]
        return d

    @classmethod
    def get_discriminator_value(cls, obj: dict[str, Optional[str]]) -> str:
        """Return the discriminator value (object type) of the data."""
        discriminator_name = obj[cls.__discriminator_property_name]
        assert discriminator_name is not None
        discriminator = cls.__discriminator_value_class_map.get(discriminator_name)
        assert discriminator is not None
        return discriminator

    @classmethod
    def from_dict(
        cls, obj: dict[str, Optional[str]]
    ) -> Union["PointOfTimeOffset", "PointOfTimeStatement", "PointOfTimeTimestamp"]:
        """Create an instance of PointOfTime from a dict."""
        object_type = cls.get_discriminator_value(obj)
        if not object_type:
            raise ValueError(
                "PointOfTime failed to lookup discriminator value from "
                + json.dumps(obj)
                + ". Discriminator property name: "
                + cls.__discriminator_property_name
                + ", mapping: "
                + json.dumps(cls.__discriminator_value_class_map)
            )
        return getattr(sys.modules[__name__], object_type).from_dict(obj)


class PointOfTimeOffset(PointOfTime):
    point_of_time_type: Annotated[str, StringConstraints(strict=True)] = "offset"


class PointOfTimeTimestamp(PointOfTime):
    point_of_time_type: Annotated[str, StringConstraints(strict=True)] = "timestamp"


class PointOfTimeStatement(PointOfTime):
    point_of_time_type: Annotated[str, StringConstraints(strict=True)] = "statement"


class TokenType(Enum):
    SESSION_TOKEN = "ST"
    EXTERNAL_SESSION_WITH_PAT = "ESPAT"


# Now that everything has been defined, let's resolve forward declarations by rebuilding a model.
Clone.model_rebuild()



from ._common import (
    Clone,
    CreateMode,
    DeleteMode,
    PointOfTime,
    PointOfTimeOffset,
    PointOfTimeStatement,
    PointOfTimeTimestamp,
)
from ._identifiers import FQN
from ._operation import PollingOperation
from ._rest_connection import RESTConnection, RESTRoot
from ._root import Root
from .logging import simple_file_logging
from .version import __version__


__all__ = [
    "Clone",
    "CreateMode",
    "DeleteMode",
    "FQN",
    "PointOfTime",
    "PointOfTimeOffset",
    "PointOfTimeStatement",
    "PointOfTimeTimestamp",
    "PollingOperation",
    "Root",
    "simple_file_logging",
    "__version__",
]



"""
Snowflake User API.

The Snowflake User API is a REST API that you can use to access, update, and perform certain action on Users in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, SecretStr, StrictBool, StrictInt, StrictStr, field_validator


class User(BaseModel):
    """A model object representing the User resource.

    Constructs an object of type User with the provided properties.

    Parameters
    __________
    name : str
        User name
    password : str, optional
        Password
    login_name : str, optional
        Login name
    display_name : str, optional
        Display name
    first_name : str, optional
        First name
    middle_name : str, optional
        Middle name
    last_name : str, optional
        Last name
    email : str, optional
        Email address
    must_change_password : bool, optional
        Does this user need to change their password (e.g., after assigning a temp password)
    disabled : bool, optional
        Has this user been disabled from the system
    days_to_expiry : int, optional
        How many days until this user expires
    mins_to_unlock : int, optional
        How many minutes until the account is unlocked after multiple failed logins
    default_warehouse : str, optional
        The default warehouse to use when this user starts a session
    default_namespace : str, optional
        The default namespace to use when this user starts a session
    default_role : str, optional
        The default role to use when this user starts a session
    default_secondary_roles : str,  default 'ALL'
        The default secondary roles of this user to use when starting a session. Only valid set values are ALL or NONE. Default is ALL after 2024-07 BCR.
    mins_to_bypass_mfa : int, optional
        How many minutes until MFA is required again
    rsa_public_key : str, optional
        RSA public key of the user
    rsa_public_key_2 : str, optional
        Second RSA public key of the user
    comment : str, optional
        Comment about the user.
    type : str, optional
        Indicates the type of user (PERSON | SERVICE | LEGACY_SERVICE)
    enable_unredacted_query_syntax_error : bool, optional
        Whether to show unredacted query syntax errors in the query history.
    network_policy : str, optional
        Specifies an existing network policy is active for the user. Otherwise, use account default.
    created_on : datetime, optional
        **Read-only:** *any user-provided value will be ignored.*
    last_successful_login : datetime, optional
        **Read-only:** *any user-provided value will be ignored.*
    expires_at : datetime, optional
        **Read-only:** *any user-provided value will be ignored.*
    locked_until : datetime, optional
        **Read-only:** *any user-provided value will be ignored.*
    has_password : bool, optional
        **Read-only:** *any user-provided value will be ignored.*
    has_rsa_public_key : bool, optional
        **Read-only:** *any user-provided value will be ignored.*
    rsa_public_key_fp : str, optional
        Fingerprint of the user's RSA public key — **Read-only:** *any user-provided value will be ignored.*
    rsa_public_key_2_fp : str, optional
        Fingerprint of the user's second RSA public key — **Read-only:** *any user-provided value will be ignored.*
    ext_authn_duo : bool, optional
        **Read-only:** *any user-provided value will be ignored.*
    ext_authn_uid : str, optional
        **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        **Read-only:** *any user-provided value will be ignored.*
    snowflake_lock : bool,  default False
        Whether the user, account, or organization is locked by Snowflake — **Read-only:** *any user-provided value will be ignored.*
    snowflake_support : bool,  default False
        Whether Snowflake Support is allowed to use the user or account — **Read-only:** *any user-provided value will be ignored.*
    mins_to_bypass_network_policy : int, optional
        Temporary bypass network policy on the user for a specified number of minutes — **Read-only:** *any user-provided value will be ignored.*
    password_last_set : datetime, optional
        **Read-only:** *any user-provided value will be ignored.*
    custom_landing_page_url : str, optional
        **Read-only:** *any user-provided value will be ignored.*
    custom_landing_page_url_flush_next_ui_load : bool,  default False
        Whether or not to flush the custom landing page of the user on next UI load — **Read-only:** *any user-provided value will be ignored.*
    """

    name: StrictStr

    password: Optional[SecretStr] = None

    login_name: Optional[StrictStr] = None

    display_name: Optional[StrictStr] = None

    first_name: Optional[StrictStr] = None

    middle_name: Optional[StrictStr] = None

    last_name: Optional[StrictStr] = None

    email: Optional[StrictStr] = None

    must_change_password: Optional[StrictBool] = None

    disabled: Optional[StrictBool] = None

    days_to_expiry: Optional[StrictInt] = None

    mins_to_unlock: Optional[StrictInt] = None

    default_warehouse: Optional[StrictStr] = None

    default_namespace: Optional[StrictStr] = None

    default_role: Optional[StrictStr] = None

    default_secondary_roles: Optional[StrictStr] = "ALL"

    mins_to_bypass_mfa: Optional[StrictInt] = None

    rsa_public_key: Optional[StrictStr] = None

    rsa_public_key_2: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    type: Optional[StrictStr] = None

    enable_unredacted_query_syntax_error: Optional[StrictBool] = None

    network_policy: Optional[StrictStr] = None

    created_on: Optional[datetime] = None

    last_successful_login: Optional[datetime] = None

    expires_at: Optional[datetime] = None

    locked_until: Optional[datetime] = None

    has_password: Optional[StrictBool] = None

    has_rsa_public_key: Optional[StrictBool] = None

    rsa_public_key_fp: Optional[StrictStr] = None

    rsa_public_key_2_fp: Optional[StrictStr] = None

    ext_authn_duo: Optional[StrictBool] = None

    ext_authn_uid: Optional[StrictStr] = None

    owner: Optional[StrictStr] = None

    snowflake_lock: Optional[StrictBool] = False

    snowflake_support: Optional[StrictBool] = False

    mins_to_bypass_network_policy: Optional[StrictInt] = None

    password_last_set: Optional[datetime] = None

    custom_landing_page_url: Optional[StrictStr] = None

    custom_landing_page_url_flush_next_ui_load: Optional[StrictBool] = False

    __properties = [
        "name",
        "password",
        "login_name",
        "display_name",
        "first_name",
        "middle_name",
        "last_name",
        "email",
        "must_change_password",
        "disabled",
        "days_to_expiry",
        "mins_to_unlock",
        "default_warehouse",
        "default_namespace",
        "default_role",
        "default_secondary_roles",
        "mins_to_bypass_mfa",
        "rsa_public_key",
        "rsa_public_key_2",
        "comment",
        "type",
        "enable_unredacted_query_syntax_error",
        "network_policy",
        "created_on",
        "last_successful_login",
        "expires_at",
        "locked_until",
        "has_password",
        "has_rsa_public_key",
        "rsa_public_key_fp",
        "rsa_public_key_2_fp",
        "ext_authn_duo",
        "ext_authn_uid",
        "owner",
        "snowflake_lock",
        "snowflake_support",
        "mins_to_bypass_network_policy",
        "password_last_set",
        "custom_landing_page_url",
        "custom_landing_page_url_flush_next_ui_load",
    ]

    @field_validator("default_secondary_roles")
    def default_secondary_roles_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("ALL", "NONE"):
            raise ValueError("must validate the enum values ('ALL','NONE')")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> User:
        """Create an instance of User from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "last_successful_login",
                    "expires_at",
                    "locked_until",
                    "has_password",
                    "has_rsa_public_key",
                    "rsa_public_key_fp",
                    "rsa_public_key_2_fp",
                    "ext_authn_duo",
                    "ext_authn_uid",
                    "owner",
                    "snowflake_lock",
                    "snowflake_support",
                    "mins_to_bypass_network_policy",
                    "password_last_set",
                    "custom_landing_page_url",
                    "custom_landing_page_url_flush_next_ui_load",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> User:
        """Create an instance of User from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return User.model_validate(obj)

        _obj = User.model_validate(
            {
                "name": obj.get("name"),
                "password": obj.get("password"),
                "login_name": obj.get("login_name"),
                "display_name": obj.get("display_name"),
                "first_name": obj.get("first_name"),
                "middle_name": obj.get("middle_name"),
                "last_name": obj.get("last_name"),
                "email": obj.get("email"),
                "must_change_password": obj.get("must_change_password"),
                "disabled": obj.get("disabled"),
                "days_to_expiry": obj.get("days_to_expiry"),
                "mins_to_unlock": obj.get("mins_to_unlock"),
                "default_warehouse": obj.get("default_warehouse"),
                "default_namespace": obj.get("default_namespace"),
                "default_role": obj.get("default_role"),
                "default_secondary_roles": obj.get("default_secondary_roles")
                if obj.get("default_secondary_roles") is not None
                else "ALL",
                "mins_to_bypass_mfa": obj.get("mins_to_bypass_mfa"),
                "rsa_public_key": obj.get("rsa_public_key"),
                "rsa_public_key_2": obj.get("rsa_public_key_2"),
                "comment": obj.get("comment"),
                "type": obj.get("type"),
                "enable_unredacted_query_syntax_error": obj.get("enable_unredacted_query_syntax_error"),
                "network_policy": obj.get("network_policy"),
                "created_on": obj.get("created_on"),
                "last_successful_login": obj.get("last_successful_login"),
                "expires_at": obj.get("expires_at"),
                "locked_until": obj.get("locked_until"),
                "has_password": obj.get("has_password"),
                "has_rsa_public_key": obj.get("has_rsa_public_key"),
                "rsa_public_key_fp": obj.get("rsa_public_key_fp"),
                "rsa_public_key_2_fp": obj.get("rsa_public_key_2_fp"),
                "ext_authn_duo": obj.get("ext_authn_duo"),
                "ext_authn_uid": obj.get("ext_authn_uid"),
                "owner": obj.get("owner"),
                "snowflake_lock": obj.get("snowflake_lock") if obj.get("snowflake_lock") is not None else False,
                "snowflake_support": obj.get("snowflake_support")
                if obj.get("snowflake_support") is not None
                else False,
                "mins_to_bypass_network_policy": obj.get("mins_to_bypass_network_policy"),
                "password_last_set": obj.get("password_last_set"),
                "custom_landing_page_url": obj.get("custom_landing_page_url"),
                "custom_landing_page_url_flush_next_ui_load": obj.get("custom_landing_page_url_flush_next_ui_load")
                if obj.get("custom_landing_page_url_flush_next_ui_load") is not None
                else False,
            }
        )

        return _obj


class UserModel:
    def __init__(
        self,
        name: str,
        # optional properties
        password: Optional[str] = None,
        login_name: Optional[str] = None,
        display_name: Optional[str] = None,
        first_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        last_name: Optional[str] = None,
        email: Optional[str] = None,
        must_change_password: Optional[bool] = None,
        disabled: Optional[bool] = None,
        days_to_expiry: Optional[int] = None,
        mins_to_unlock: Optional[int] = None,
        default_warehouse: Optional[str] = None,
        default_namespace: Optional[str] = None,
        default_role: Optional[str] = None,
        default_secondary_roles: Optional[str] = "ALL",
        mins_to_bypass_mfa: Optional[int] = None,
        rsa_public_key: Optional[str] = None,
        rsa_public_key_2: Optional[str] = None,
        comment: Optional[str] = None,
        type: Optional[str] = None,
        enable_unredacted_query_syntax_error: Optional[bool] = None,
        network_policy: Optional[str] = None,
        created_on: Optional[datetime] = None,
        last_successful_login: Optional[datetime] = None,
        expires_at: Optional[datetime] = None,
        locked_until: Optional[datetime] = None,
        has_password: Optional[bool] = None,
        has_rsa_public_key: Optional[bool] = None,
        rsa_public_key_fp: Optional[str] = None,
        rsa_public_key_2_fp: Optional[str] = None,
        ext_authn_duo: Optional[bool] = None,
        ext_authn_uid: Optional[str] = None,
        owner: Optional[str] = None,
        snowflake_lock: Optional[bool] = False,
        snowflake_support: Optional[bool] = False,
        mins_to_bypass_network_policy: Optional[int] = None,
        password_last_set: Optional[datetime] = None,
        custom_landing_page_url: Optional[str] = None,
        custom_landing_page_url_flush_next_ui_load: Optional[bool] = False,
    ):
        """A model object representing the User resource.

        Constructs an object of type User with the provided properties.

        Parameters
        __________
        name : str
            User name
        password : str, optional
            Password
        login_name : str, optional
            Login name
        display_name : str, optional
            Display name
        first_name : str, optional
            First name
        middle_name : str, optional
            Middle name
        last_name : str, optional
            Last name
        email : str, optional
            Email address
        must_change_password : bool, optional
            Does this user need to change their password (e.g., after assigning a temp password)
        disabled : bool, optional
            Has this user been disabled from the system
        days_to_expiry : int, optional
            How many days until this user expires
        mins_to_unlock : int, optional
            How many minutes until the account is unlocked after multiple failed logins
        default_warehouse : str, optional
            The default warehouse to use when this user starts a session
        default_namespace : str, optional
            The default namespace to use when this user starts a session
        default_role : str, optional
            The default role to use when this user starts a session
        default_secondary_roles : str,  default 'ALL'
            The default secondary roles of this user to use when starting a session. Only valid set values are ALL or NONE. Default is ALL after 2024-07 BCR.
        mins_to_bypass_mfa : int, optional
            How many minutes until MFA is required again
        rsa_public_key : str, optional
            RSA public key of the user
        rsa_public_key_2 : str, optional
            Second RSA public key of the user
        comment : str, optional
            Comment about the user.
        type : str, optional
            Indicates the type of user (PERSON | SERVICE | LEGACY_SERVICE)
        enable_unredacted_query_syntax_error : bool, optional
            Whether to show unredacted query syntax errors in the query history.
        network_policy : str, optional
            Specifies an existing network policy is active for the user. Otherwise, use account default.
        created_on : datetime, optional

        last_successful_login : datetime, optional

        expires_at : datetime, optional

        locked_until : datetime, optional

        has_password : bool, optional

        has_rsa_public_key : bool, optional

        rsa_public_key_fp : str, optional
            Fingerprint of the user's RSA public key
        rsa_public_key_2_fp : str, optional
            Fingerprint of the user's second RSA public key
        ext_authn_duo : bool, optional

        ext_authn_uid : str, optional

        owner : str, optional

        snowflake_lock : bool,  default False
            Whether the user, account, or organization is locked by Snowflake.
        snowflake_support : bool,  default False
            Whether Snowflake Support is allowed to use the user or account
        mins_to_bypass_network_policy : int, optional
            Temporary bypass network policy on the user for a specified number of minutes
        password_last_set : datetime, optional

        custom_landing_page_url : str, optional

        custom_landing_page_url_flush_next_ui_load : bool,  default False
            Whether or not to flush the custom landing page of the user on next UI load
        """
        self.name = name
        self.password = password
        self.login_name = login_name
        self.display_name = display_name
        self.first_name = first_name
        self.middle_name = middle_name
        self.last_name = last_name
        self.email = email
        self.must_change_password = must_change_password
        self.disabled = disabled
        self.days_to_expiry = days_to_expiry
        self.mins_to_unlock = mins_to_unlock
        self.default_warehouse = default_warehouse
        self.default_namespace = default_namespace
        self.default_role = default_role
        self.default_secondary_roles = default_secondary_roles
        self.mins_to_bypass_mfa = mins_to_bypass_mfa
        self.rsa_public_key = rsa_public_key
        self.rsa_public_key_2 = rsa_public_key_2
        self.comment = comment
        self.type = type
        self.enable_unredacted_query_syntax_error = enable_unredacted_query_syntax_error
        self.network_policy = network_policy
        self.created_on = created_on
        self.last_successful_login = last_successful_login
        self.expires_at = expires_at
        self.locked_until = locked_until
        self.has_password = has_password
        self.has_rsa_public_key = has_rsa_public_key
        self.rsa_public_key_fp = rsa_public_key_fp
        self.rsa_public_key_2_fp = rsa_public_key_2_fp
        self.ext_authn_duo = ext_authn_duo
        self.ext_authn_uid = ext_authn_uid
        self.owner = owner
        self.snowflake_lock = snowflake_lock
        self.snowflake_support = snowflake_support
        self.mins_to_bypass_network_policy = mins_to_bypass_network_policy
        self.password_last_set = password_last_set
        self.custom_landing_page_url = custom_landing_page_url
        self.custom_landing_page_url_flush_next_ui_load = custom_landing_page_url_flush_next_ui_load

    __properties = [
        "name",
        "password",
        "login_name",
        "display_name",
        "first_name",
        "middle_name",
        "last_name",
        "email",
        "must_change_password",
        "disabled",
        "days_to_expiry",
        "mins_to_unlock",
        "default_warehouse",
        "default_namespace",
        "default_role",
        "default_secondary_roles",
        "mins_to_bypass_mfa",
        "rsa_public_key",
        "rsa_public_key_2",
        "comment",
        "type",
        "enable_unredacted_query_syntax_error",
        "network_policy",
        "created_on",
        "last_successful_login",
        "expires_at",
        "locked_until",
        "has_password",
        "has_rsa_public_key",
        "rsa_public_key_fp",
        "rsa_public_key_2_fp",
        "ext_authn_duo",
        "ext_authn_uid",
        "owner",
        "snowflake_lock",
        "snowflake_support",
        "mins_to_bypass_network_policy",
        "password_last_set",
        "custom_landing_page_url",
        "custom_landing_page_url_flush_next_ui_load",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return User(
            name=self.name,
            password=self.password,
            login_name=self.login_name,
            display_name=self.display_name,
            first_name=self.first_name,
            middle_name=self.middle_name,
            last_name=self.last_name,
            email=self.email,
            must_change_password=self.must_change_password,
            disabled=self.disabled,
            days_to_expiry=self.days_to_expiry,
            mins_to_unlock=self.mins_to_unlock,
            default_warehouse=self.default_warehouse,
            default_namespace=self.default_namespace,
            default_role=self.default_role,
            default_secondary_roles=self.default_secondary_roles,
            mins_to_bypass_mfa=self.mins_to_bypass_mfa,
            rsa_public_key=self.rsa_public_key,
            rsa_public_key_2=self.rsa_public_key_2,
            comment=self.comment,
            type=self.type,
            enable_unredacted_query_syntax_error=self.enable_unredacted_query_syntax_error,
            network_policy=self.network_policy,
            created_on=self.created_on,
            last_successful_login=self.last_successful_login,
            expires_at=self.expires_at,
            locked_until=self.locked_until,
            has_password=self.has_password,
            has_rsa_public_key=self.has_rsa_public_key,
            rsa_public_key_fp=self.rsa_public_key_fp,
            rsa_public_key_2_fp=self.rsa_public_key_2_fp,
            ext_authn_duo=self.ext_authn_duo,
            ext_authn_uid=self.ext_authn_uid,
            owner=self.owner,
            snowflake_lock=self.snowflake_lock,
            snowflake_support=self.snowflake_support,
            mins_to_bypass_network_policy=self.mins_to_bypass_network_policy,
            password_last_set=self.password_last_set,
            custom_landing_page_url=self.custom_landing_page_url,
            custom_landing_page_url_flush_next_ui_load=self.custom_landing_page_url_flush_next_ui_load,
        )

    @classmethod
    def _from_model(cls, model) -> UserModel:
        return UserModel(
            name=model.name,
            password=model.password,
            login_name=model.login_name,
            display_name=model.display_name,
            first_name=model.first_name,
            middle_name=model.middle_name,
            last_name=model.last_name,
            email=model.email,
            must_change_password=model.must_change_password,
            disabled=model.disabled,
            days_to_expiry=model.days_to_expiry,
            mins_to_unlock=model.mins_to_unlock,
            default_warehouse=model.default_warehouse,
            default_namespace=model.default_namespace,
            default_role=model.default_role,
            default_secondary_roles=model.default_secondary_roles,
            mins_to_bypass_mfa=model.mins_to_bypass_mfa,
            rsa_public_key=model.rsa_public_key,
            rsa_public_key_2=model.rsa_public_key_2,
            comment=model.comment,
            type=model.type,
            enable_unredacted_query_syntax_error=model.enable_unredacted_query_syntax_error,
            network_policy=model.network_policy,
            created_on=model.created_on,
            last_successful_login=model.last_successful_login,
            expires_at=model.expires_at,
            locked_until=model.locked_until,
            has_password=model.has_password,
            has_rsa_public_key=model.has_rsa_public_key,
            rsa_public_key_fp=model.rsa_public_key_fp,
            rsa_public_key_2_fp=model.rsa_public_key_2_fp,
            ext_authn_duo=model.ext_authn_duo,
            ext_authn_uid=model.ext_authn_uid,
            owner=model.owner,
            snowflake_lock=model.snowflake_lock,
            snowflake_support=model.snowflake_support,
            mins_to_bypass_network_policy=model.mins_to_bypass_network_policy,
            password_last_set=model.password_last_set,
            custom_landing_page_url=model.custom_landing_page_url,
            custom_landing_page_url_flush_next_ui_load=model.custom_landing_page_url_flush_next_ui_load,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a User.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the User object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> UserModel:
        """Create an instance of User from a dict.

        This method constructs a User object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        User
            A User object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(User.from_dict(obj))


User._model_class = UserModel



from io import BytesIO
from textwrap import dedent

import pytest as pytest

from snowflake.core.service import ServiceSpecStageFile
from snowflake.core.stage import Stage

from ..utils import random_string


@pytest.fixture(scope="module")
def service_spec_file_on_stage(stages, imagerepo, session) -> ServiceSpecStageFile:
    stage_name = random_string(5, "test_stage_")
    stage = Stage(name=stage_name, comment="created by service_spec_file_on_stage")
    temp_stage = stages.create(stage)
    try:
        spec_file = "spec.yaml"
        stage_file = f"@{temp_stage.name}"
        spec = f"{stage_file}/{spec_file}"
        session.file.put_stream(
            BytesIO(
                dedent(f"""
                        spec:
                          containers:
                          - name: hello-world
                            image: {imagerepo}/hello-world:latest
                        """).encode()
            ),
            spec,
        )

        yield ServiceSpecStageFile(stage=temp_stage.name, spec_file=spec_file)
    finally:
        temp_stage.drop()



import pytest

from snowflake.core.exceptions import APIError, ConflictError
from tests.integ.external_volume.utils import get_aws_external_volume_template


pytestmark = [pytest.mark.internal_only]


def test_undrop_external_volume(external_volumes, setup_credentials_fixture):
    try:
        external_volume_to_drop = []
        # happy path aws
        external_volume = get_aws_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]

        # happy drop
        external_volume_handler.drop()

        # happy undrop
        external_volume_handler.undrop()

        # undrop when it is not dropped
        with pytest.raises(ConflictError):
            external_volume_handler.undrop()

        # undrop random external volume
        with pytest.raises(APIError):
            external_volumes["RANDOM"].undrop()

    finally:
        for external_volume_handler in external_volume_to_drop:
            external_volume_handler.drop(if_exists=True)



"""Manages Snowpark Container Services.

Example:
    >>> new_service_def = Service(
    ...     name="MYSERVICE",
    ...     compute_pool="MYCOMPUTEPOOL",
    ...     spec="@~/myservice_spec.yml",
    ...     min_instances=1,
    ...     max_instances=1,
    ... )
    >>> services = root.databases["MYDB"].schemas["MYSCHEMA"].services
    >>> myservice = services.create(new_service_def)
    >>> myservice_snapshot = myservice.fetch()
    >>> service_data = services.iter(like="%SERVICE")
    >>> myservice.suspend()
    >>> myservice.resume()
    >>> service_status = myservice.get_service_status()
    >>> logs = myservice.get_service_logs()
    >>> myservice.delete()
    >>> an_existing_service = services["an_existing_service"]
    >>> an_existing_service.suspend()

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from snowflake.core.service._generated.models import ServiceSpec as ServiceSpecification

from ._service import (
    GrantOf,
    JobService,
    Service,
    ServiceCollection,
    ServiceContainer,
    ServiceEndpoint,
    ServiceInstance,
    ServiceResource,
    ServiceRole,
    ServiceRoleGrantTo,
    ServiceSpec,
    ServiceSpecInlineText,
    ServiceSpecStageFile,
)


__all__ = [
    "GrantOf",
    "Service",
    "ServiceCollection",
    "ServiceResource",
    "ServiceSpec",
    "ServiceSpecification",
    "ServiceSpecInlineText",
    "ServiceSpecStageFile",
    "JobService",
    "ServiceContainer",
    "ServiceEndpoint",
    "ServiceInstance",
    "ServiceRole",
    "ServiceRoleGrantTo",
]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from snowflake.core.compute_pool import ComputePool
from snowflake.core.exceptions import NotFoundError

from ..utils import random_string


pytestmark = [pytest.mark.skip_gov]


def test_drop(compute_pools, instance_family):
    cp_name = random_string(5, "test_cp_")
    test_cp = ComputePool(name=cp_name, instance_family=instance_family, min_nodes=1, max_nodes=1, auto_resume=False)
    compute_pools.create(test_cp, initially_suspended=True)
    compute_pools[test_cp.name].drop()
    with pytest.raises(NotFoundError):
        compute_pools[test_cp.name].fetch()

    # create again the same cp and drop
    compute_pools.create(test_cp, initially_suspended=True)
    compute_pools[test_cp.name].drop()

    # drop when not exists
    with pytest.raises(NotFoundError):
        compute_pools[test_cp.name].drop()

    # drop when not exists with if_exists
    compute_pools[test_cp.name].drop(if_exists=True)
